# -*- coding: utf-8 -*-

# if False:
#    import os
#    os.environ['UTOOL_NOCNN'] = 'True'
import logging
import six
import utool as ut
import uuid  # NOQA
from wbia.control import controller_inject
import wbia.constants as const

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)
register_api = controller_inject.get_wbia_flask_api(__name__)


def ensure_simple_server(port=5832):
    r"""
    CommandLine:
        python -m wbia.web.apis_engine --exec-ensure_simple_server
        python -m utool.util_web --exec-start_simple_webserver

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> result = ensure_simple_server()
        >>> print(result)
    """
    if ut.is_local_port_open(port):
        bgserver = ut.spawn_background_process(ut.start_simple_webserver, port=port)
        return bgserver
    else:
        bgserver = ut.DynStruct()
        bgserver.terminate2 = lambda: None
        logger.info('server is running elsewhere')
    return bgserver


def ensure_uuid_list(list_):
    if list_ is not None and len(list_) > 0 and isinstance(list_[0], six.string_types):
        list_ = list(map(uuid.UUID, list_))
    return list_


def web_check_annot_uuids_with_names(annot_uuid_list, name_list):
    assert None not in [annot_uuid_list, name_list]
    assert len(annot_uuid_list) == len(name_list)

    bad_uuid_set = set([])

    annot_dict = {}
    zipped = zip(annot_uuid_list, name_list)
    for index, (annot_uuid, name) in enumerate(zipped):
        if not isinstance(annot_uuid, (uuid.UUID, six.string_types)):
            raise ValueError(
                'Received UUID %r that is not a UUID or string at index %s'
                % (annot_uuid, index)
            )

        if annot_uuid not in annot_dict:
            annot_dict[annot_uuid] = set([])

        if name not in ['', const.UNKNOWN, None]:
            annot_dict[annot_uuid].add(name)

        if len(annot_dict[annot_uuid]) > 1:
            bad_uuid_set.add(annot_uuid)

    if len(bad_uuid_set) > 0:
        bad_dict = {bad_uuid: annot_dict[bad_uuid] for bad_uuid in bad_uuid_set}
        raise controller_inject.WebMultipleNamedDuplicateException(bad_dict)

    annot_uuid_list_ = []
    name_list_ = []
    for annot_uuid in annot_dict:
        names = list(annot_dict.get(annot_uuid, []))
        if len(names) == 0:
            name = const.UNKNOWN
        elif len(names) == 1:
            name = names[0]
        else:
            raise RuntimeError('Sanity check failed in web_check_annot_uuids_with_names')

        annot_uuid_list_.append(annot_uuid)
        name_list_.append(name)

    return annot_uuid_list_, name_list_


@register_ibs_method
@register_api('/api/engine/uuid/check/', methods=['GET', 'POST'])
def web_check_uuids(ibs, image_uuid_list=[], qannot_uuid_list=[], dannot_uuid_list=[]):
    r"""
    Args:
        ibs (wbia.IBEISController):  image analysis api
        image_uuid_list (list): (default = [])
        qannot_uuid_list (list): (default = [])
        dannot_uuid_list (list): (default = [])

    CommandLine:
        python -m wbia.web.apis_engine --exec-web_check_uuids --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> image_uuid_list = []
        >>> qannot_uuid_list = ibs.get_annot_uuids([1, 1, 2, 3, 2, 4])
        >>> dannot_uuid_list = ibs.get_annot_uuids([1, 2, 3])
        >>> try:
        >>>     web_check_uuids(ibs, image_uuid_list, qannot_uuid_list,
        >>>                     dannot_uuid_list)
        >>> except controller_inject.WebDuplicateUUIDException:
        >>>     pass
        >>> else:
        >>>     raise AssertionError('Should have gotten WebDuplicateUUIDException')
        >>> try:
        >>>     web_check_uuids(ibs, [1, 2, 3], qannot_uuid_list,
        >>>                     dannot_uuid_list)
        >>> except controller_inject.WebMissingUUIDException as ex:
        >>>     pass
        >>> else:
        >>>     raise AssertionError('Should have gotten WebMissingUUIDException')
        >>> print('Successfully reported errors')
    """
    import uuid

    # Unique list
    if qannot_uuid_list is None:
        qannot_uuid_list = []
    if dannot_uuid_list is None:
        dannot_uuid_list = []

    annot_uuid_list = qannot_uuid_list + dannot_uuid_list

    # UUID parse check
    invalid_image_uuid_list = []
    for index, image_uuid in enumerate(image_uuid_list):
        try:
            assert isinstance(image_uuid, uuid.UUID)
        except Exception:
            value = (
                index,
                image_uuid,
            )
            invalid_image_uuid_list.append(value)

    invalid_annot_uuid_list = []
    for index, annot_uuid in enumerate(annot_uuid_list):
        try:
            assert isinstance(annot_uuid, uuid.UUID)
        except Exception:
            value = (
                index,
                annot_uuid,
            )
            invalid_annot_uuid_list.append(value)

    if len(invalid_image_uuid_list) > 0 or len(invalid_annot_uuid_list) > 0:
        kwargs = {
            'invalid_image_uuid_list': invalid_image_uuid_list,
            'invalid_annot_uuid_list': invalid_annot_uuid_list,
        }
        raise controller_inject.WebInvalidUUIDException(**kwargs)

    image_uuid_list = list(set(image_uuid_list))
    annot_uuid_list = list(set(annot_uuid_list))
    # Check for all annot UUIDs exist
    missing_image_uuid_list = ibs.get_image_missing_uuid(image_uuid_list)
    missing_annot_uuid_list = ibs.get_annot_missing_uuid(annot_uuid_list)
    if len(missing_image_uuid_list) > 0 or len(missing_annot_uuid_list) > 0:
        kwargs = {
            'missing_image_uuid_list': missing_image_uuid_list,
            'missing_annot_uuid_list': missing_annot_uuid_list,
        }
        raise controller_inject.WebMissingUUIDException(**kwargs)
    ddup_pos_map = ut.find_duplicate_items(qannot_uuid_list)
    qdup_pos_map = ut.find_duplicate_items(dannot_uuid_list)
    if len(ddup_pos_map) + len(qdup_pos_map) > 0:
        raise controller_inject.WebDuplicateUUIDException(qdup_pos_map, qdup_pos_map)


@register_ibs_method
@register_api('/api/engine/query/annot/rowid/', methods=['GET', 'POST'])
def start_identify_annots(
    ibs,
    qannot_uuid_list,
    dannot_uuid_list=None,
    pipecfg={},
    callback_url=None,
    callback_method=None,
    lane='slow',
):
    r"""
    REST:
        Method: GET
        URL: /api/engine/query/annot/rowid/

    Args:
        qannot_uuid_list (list) : specifies the query annotations to
            identify.
        dannot_uuid_list (list) : specifies the annotations that the
            algorithm is allowed to use for identification.  If not
            specified all annotations are used.   (default=None)
        pipecfg (dict) : dictionary of pipeline configuration arguments
            (default=None)

    CommandLine:
        # Run as main process
        python -m wbia.web.apis_engine --exec-start_identify_annots:0
        # Run using server process
        python -m wbia.web.apis_engine --exec-start_identify_annots:1

        # Split into multiple processes
        python -m wbia.web.apis_engine --main --bg
        python -m wbia.web.apis_engine --exec-start_identify_annots:1 --fg

        python -m wbia.web.apis_engine --exec-start_identify_annots:1 --domain http://52.33.105.88

        python -m wbia.web.apis_engine --exec-start_identify_annots:1 --duuids=[]
        python -m wbia.web.apis_engine --exec-start_identify_annots:1 --domain http://52.33.105.88 --duuids=03a17411-c226-c960-d180-9fafef88c880


    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> from wbia.web import apis_engine
        >>> import wbia
        >>> ibs, qaids, daids = wbia.testdata_expanded_aids(
        >>>     defaultdb='PZ_MTEST', a=['default:qsize=2,dsize=10'])
        >>> qannot_uuid_list = ibs.get_annot_uuids(qaids)
        >>> dannot_uuid_list = ibs.get_annot_uuids(daids)
        >>> pipecfg = {}
        >>> ibs.initialize_job_manager()
        >>> jobid = ibs.start_identify_annots(qannot_uuid_list, dannot_uuid_list, pipecfg)
        >>> result = ibs.wait_for_job_result(jobid, timeout=None, freq=2)
        >>> print(result)
        >>> import utool as ut
        >>> #print(ut.to_json(result))
        >>> ibs.close_job_manager()

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')  # , domain='http://52.33.105.88')
        >>> aids = ibs.get_valid_aids()[0:2]
        >>> qaids = aids[0:1]
        >>> daids = aids
        >>> query_config_dict = {
        >>>     #'pipeline_root' : 'BC_DTW'
        >>> }
        >>> qreq_ = ibs.new_query_request(qaids, daids, cfgdict=query_config_dict)
        >>> cm_list = qreq_.execute()

    Example:
        >>> # xdoctest: +REQUIRES(--web-tests)
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> import wbia
        >>> with wbia.opendb_bg_web('testdb1', managed=True) as web_ibs:  # , domain='http://52.33.105.88')
        ...     aids = web_ibs.send_wbia_request('/api/annot/', 'get')[0:2]
        ...     uuid_list = web_ibs.send_wbia_request('/api/annot/uuid/', type_='get', aid_list=aids)
        ...     quuid_list = ut.get_argval('--quuids', type_=list, default=uuid_list)
        ...     duuid_list = ut.get_argval('--duuids', type_=list, default=uuid_list)
        ...     data = dict(
        ...         qannot_uuid_list=quuid_list, dannot_uuid_list=duuid_list,
        ...         pipecfg={},
        ...         callback_url='http://127.0.1.1:5832'
        ...     )
        ...     # Start callback server
        ...     bgserver = ensure_simple_server()
        ...     # --
        ...     jobid = web_ibs.send_wbia_request('/api/engine/query/annot/rowid/', **data)
        ...     status_response = web_ibs.wait_for_results(jobid, delays=[1, 5, 30])
        ...     print('status_response = %s' % (status_response,))
        ...     result_response = web_ibs.read_engine_results(jobid)
        ...     print('result_response = %s' % (result_response,))
        ...     cm_dict = result_response['json_result'][0]
        ...     print('Finished test')
        ...     bgserver.terminate2()
        Waiting for server to be up. count=0
        ...
        status_response = {'status': 'ok', 'jobid': '...', 'jobstatus': 'completed'}
        result_response = ...
        Finished test

    Ignore:
        qaids = daids = ibs.get_valid_aids()
        jobid = ibs.start_identify_annots(**payload)
    """
    # Check UUIDs
    ibs.web_check_uuids([], qannot_uuid_list, dannot_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    qannot_uuid_list = ensure_uuid_list(qannot_uuid_list)
    dannot_uuid_list = ensure_uuid_list(dannot_uuid_list)

    qaid_list = ibs.get_annot_aids_from_uuid(qannot_uuid_list)
    if dannot_uuid_list is None:
        daid_list = ibs.get_valid_aids()
    else:
        if len(dannot_uuid_list) == 1 and dannot_uuid_list[0] is None:
            # VERY HACK
            daid_list = ibs.get_valid_aids()
        else:
            daid_list = ibs.get_annot_aids_from_uuid(dannot_uuid_list)

    ibs.assert_valid_aids(
        qaid_list, msg='error in start_identify qaids', auuid_list=qannot_uuid_list
    )
    ibs.assert_valid_aids(
        daid_list, msg='error in start_identify daids', auuid_list=dannot_uuid_list
    )
    args = (qaid_list, daid_list, pipecfg)
    jobid = ibs.job_manager.jobiface.queue_job(
        'query_chips_simple_dict', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/api/engine/query/graph/complete/', methods=['GET', 'POST'])
def start_identify_annots_query_complete(
    ibs,
    annot_uuid_list=None,
    annot_name_list=None,
    matching_state_list=[],
    query_config_dict={},
    k=5,
    echo_query_params=True,
    callback_url=None,
    callback_method=None,
    lane='slow',
):
    r"""
    REST:
        Method: GET
        URL: /api/engine/query/complete/

    Args:
        annot_uuid_list (list) : specifies the query annotations to
            identify.
        annot_name_list (list) : specifies the query annotation names
        matching_state_list (list of tuple) : the list of matching state
            3-tuples corresponding to the query_annot_uuid_list (default=None)
        query_config_dict (dict) : dictionary of algorithmic configuration
            arguments.  (default=None)
        echo_query_params (bool) : flag for if to return the original query
            parameters with the result
    """
    # Check inputs
    if annot_uuid_list is not None and annot_name_list is not None:
        assert len(annot_uuid_list) == len(annot_name_list)

    # Check UUIDs
    if annot_name_list is not None:
        vals = web_check_annot_uuids_with_names(annot_uuid_list, annot_name_list)
        annot_uuid_list, annot_name_list = vals
    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)

    annot_uuid_list = ensure_uuid_list(annot_uuid_list)

    # Ensure annotations
    if annot_uuid_list is None or (
        len(annot_uuid_list) == 1 and annot_uuid_list[0] is None
    ):
        aid_list = ibs.get_valid_aids()
    else:
        aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    # Ensure names
    # FIXME: THE QREQ STRUCTURE SHOULD HANDLE NAMES.
    if annot_name_list is not None:
        # Set names for query annotations
        nid_list = ibs.add_names(annot_name_list)
        ibs.set_annot_name_rowids(aid_list, nid_list)

    ibs.assert_valid_aids(
        aid_list, msg='error in start_identify qaids', auuid_list=annot_uuid_list
    )

    args = (
        aid_list,
        query_config_dict,
        k,
    )
    jobid = ibs.job_manager.jobiface.queue_job(
        'query_chips_graph_complete', callback_url, callback_method, lane, *args
    )
    return jobid


@register_ibs_method
@register_api('/api/engine/query/graph/', methods=['GET', 'POST'])
def start_identify_annots_query(
    ibs,
    query_annot_uuid_list=None,
    # query_annot_name_uuid_list=None,
    query_annot_name_list=None,
    database_annot_uuid_list=None,
    # database_annot_name_uuid_list=None,
    database_annot_name_list=None,
    matching_state_list=[],
    query_config_dict={},
    echo_query_params=True,
    include_qaid_in_daids=True,
    callback_url=None,
    callback_method=None,
    lane='slow',
):
    r"""
    REST:
        Method: GET
        URL: /api/engine/query/graph/

    Args:
        query_annot_uuid_list (list) : specifies the query annotations to
            identify.
        query_annot_name_list (list) : specifies the query annotation names
        database_annot_uuid_list (list) : specifies the annotations that the
            algorithm is allowed to use for identification.  If not
            specified all annotations are used.   (default=None)
        database_annot_name_list (list) : specifies the database annotation
            names (default=None)
        matching_state_list (list of tuple) : the list of matching state
            3-tuples corresponding to the query_annot_uuid_list (default=None)
        query_config_dict (dict) : dictionary of algorithmic configuration
            arguments.  (default=None)
        echo_query_params (bool) : flag for if to return the original query
            parameters with the result

    CommandLine:
        # Normal mode
        python -m wbia.web.apis_engine start_identify_annots_query
        # Split mode
        wbia --web
        python -m wbia.web.apis_engine start_identify_annots_query --show --domain=localhost

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> import wbia
        >>> #domain = 'localhost'
        >>> domain = None
        >>> with wbia.opendb_bg_web('testdb1', domain=domain, managed=True) as web_ibs:  # , domain='http://52.33.105.88')
        ...     aids = web_ibs.send_wbia_request('/api/annot/', 'get')[0:3]
        ...     uuid_list = web_ibs.send_wbia_request('/api/annot/uuid/', type_='get', aid_list=aids)
        ...     quuid_list = ut.get_argval('--quuids', type_=list, default=uuid_list)[0:1]
        ...     duuid_list = ut.get_argval('--duuids', type_=list, default=uuid_list)
        ...     query_config_dict = {
        ...        #'pipeline_root' : 'BC_DTW'
        ...     }
        ...     data = dict(
        ...         query_annot_uuid_list=quuid_list, database_annot_uuid_list=duuid_list,
        ...         query_config_dict=query_config_dict,
        ...     )
        ...     jobid = web_ibs.send_wbia_request('/api/engine/query/graph/', **data)
        ...     print('jobid = %r' % (jobid,))
        ...     status_response = web_ibs.wait_for_results(jobid)
        ...     result_response = web_ibs.read_engine_results(jobid)
        ...     print('result_response = %s' % (ut.repr3(result_response),))
        ...     inference_result = result_response['json_result']
        ...     if isinstance(inference_result, six.string_types):
        ...        print(inference_result)
        ...     cm_dict = inference_result['cm_dict']
        ...     quuid = quuid_list[0]
        ...     cm = cm_dict[str(quuid)]

    """
    valid_states = {
        'match': ['matched'],  # ['match', 'matched'],
        'nomatch': [
            'notmatched',
            'nonmatch',
        ],  # ['nomatch', 'notmatched', 'nonmatched', 'notmatch', 'non-match', 'not-match'],
        'notcomp': ['notcomparable'],
    }
    prefered_states = ut.take_column(valid_states.values(), 0)
    flat_states = ut.flatten(valid_states.values())

    def sanitize(state):
        state = state.strip().lower()
        state = ''.join(state.split())
        assert (
            state in flat_states
        ), 'matching_state_list has unrecognized states. Should be one of %r' % (
            prefered_states,
        )
        return state

    # HACK
    # if query_annot_uuid_list is None:
    #     if True:
    #         query_annot_uuid_list = []
    #     else:
    #         query_annot_uuid_list = ibs.get_annot_uuids(ibs.get_valid_aids()[0:1])

    dname_list = database_annot_name_list
    qname_list = query_annot_name_list

    # Check inputs
    assert (
        len(query_annot_uuid_list) == 1
    ), 'Can only identify one query annotation at a time. Got %d ' % (
        len(query_annot_uuid_list),
    )
    assert (
        database_annot_uuid_list is None or len(database_annot_uuid_list) > 0
    ), 'Cannot specify an empty database_annot_uuid_list (e.g. []), specify or use None instead'

    if qname_list is not None:
        assert len(query_annot_uuid_list) == len(qname_list)
    if database_annot_uuid_list is not None and dname_list is not None:
        assert len(database_annot_uuid_list) == len(dname_list)

    proot = query_config_dict.get('pipeline_root', 'vsmany')
    proot = query_config_dict.get('proot', proot)
    if proot.lower() in (
        'curvrankdorsal',
        'curvrankfinfindrhybriddorsal',
        'curvrankfluke',
    ):
        curvrank_daily_tag = query_config_dict.get('curvrank_daily_tag', '')
        if len(curvrank_daily_tag) > 144:
            message = 'The curvrank_daily_tag cannot have more than 128 characters, please shorten and try again.'
            key = 'query_config_dict["curvrank_daily_tag"]'
            value = curvrank_daily_tag
            raise controller_inject.WebInvalidInput(message, key, value)
    else:
        logger.info(ut.repr3(query_config_dict))
        pop_key_list = ['curvrank_daily_tag']
        for pop_key in pop_key_list:
            logger.info('Popping irrelevant key for config: %r' % (pop_key,))
            query_config_dict.pop(pop_key, None)

    # Check UUIDs
    if dname_list is not None:
        vals = web_check_annot_uuids_with_names(database_annot_uuid_list, dname_list)
        database_annot_uuid_list, dname_list = vals
    ibs.web_check_uuids([], query_annot_uuid_list, database_annot_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)

    qannot_uuid_list = ensure_uuid_list(query_annot_uuid_list)
    dannot_uuid_list = ensure_uuid_list(database_annot_uuid_list)

    # Ensure annotations
    qaid_list = ibs.get_annot_aids_from_uuid(qannot_uuid_list)

    if dannot_uuid_list is None or (
        len(dannot_uuid_list) == 1 and dannot_uuid_list[0] is None
    ):
        qaid = qaid_list[0]
        species = ibs.get_annot_species_texts(qaid)
        assert (
            species != const.UNKNOWN
        ), 'Cannot query an annotation with an empty (unset) species against an unspecified database_annot_uuid_list'
        daid_list = ibs.get_valid_aids()
        daid_list = ibs.filter_annotation_set(daid_list, species=species)
    else:
        daid_list = ibs.get_annot_aids_from_uuid(dannot_uuid_list)

    # Ensure that the daid_list is non-empty after filtering out qaid_list
    daid_remaining_set = set(daid_list) - set(qaid_list)
    if len(daid_remaining_set) == 0:
        raise controller_inject.WebInvalidMatchException(qaid_list, daid_list)

    # Ensure names
    # FIXME: THE QREQ STRUCTURE SHOULD HANDLE NAMES.
    if qname_list is not None:
        # Set names for query annotations
        qnid_list = ibs.add_names(qname_list)
        ibs.set_annot_name_rowids(qaid_list, qnid_list)

    if dname_list is not None:
        # Set names for database annotations
        dnid_list = ibs.add_names(dname_list)
        ibs.set_annot_name_rowids(daid_list, dnid_list)

    # For caching reasons, let's add the qaids to the daids
    if include_qaid_in_daids:
        daid_list = daid_list + qaid_list
        daid_list = sorted(daid_list)

    # Convert annot UUIDs to aids for matching_state_list into user_feedback for query
    state_list = map(sanitize, ut.take_column(matching_state_list, 2))
    # {'aid1': [1], 'aid2': [2], 'p_match': [1.0], 'p_nomatch': [0.0], 'p_notcomp': [0.0]}
    user_feedback = {
        'aid1': ibs.get_annot_aids_from_uuid(ut.take_column(matching_state_list, 0)),
        'aid2': ibs.get_annot_aids_from_uuid(ut.take_column(matching_state_list, 1)),
        'p_match': [
            1.0 if state in valid_states['match'] else 0.0 for state in state_list
        ],
        'p_nomatch': [
            1.0 if state in valid_states['nomatch'] else 0.0 for state in state_list
        ],
        'p_notcomp': [
            1.0 if state in valid_states['notcomp'] else 0.0 for state in state_list
        ],
    }

    ibs.assert_valid_aids(
        qaid_list, msg='error in start_identify qaids', auuid_list=qannot_uuid_list
    )
    ibs.assert_valid_aids(
        daid_list, msg='error in start_identify daids', auuid_list=dannot_uuid_list
    )
    args = (qaid_list, daid_list, user_feedback, query_config_dict, echo_query_params)
    jobid = ibs.job_manager.jobiface.queue_job(
        'query_chips_graph', callback_url, callback_method, lane, *args
    )
    return jobid


@register_ibs_method
@register_api('/api/engine/wic/cnn/', methods=['POST'])
def start_wic_image(
    ibs, image_uuid_list, callback_url=None, callback_method=None, lane='fast', **kwargs
):
    """
    REST:
        Method: GET
        URL: /api/engine/wic/cnn/

    Args:
        image_uuid_list (list) : list of image uuids to detect on.
        callback_url (url) : url that will be called when detection succeeds or fails
    """
    # Check UUIDs
    ibs.web_check_uuids(image_uuid_list=image_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    image_uuid_list = ensure_uuid_list(image_uuid_list)
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    args = (
        gid_list,
        kwargs,
    )
    jobid = ibs.job_manager.jobiface.queue_job(
        'wic_cnn_json', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/api/engine/detect/cnn/yolo/', methods=['POST'])
def start_detect_image_yolo(
    ibs, image_uuid_list, callback_url=None, callback_method=None, lane='fast', **kwargs
):
    """
    REST:
        Method: GET
        URL: /api/engine/detect/cnn/yolo/

    Args:
        image_uuid_list (list) : list of image uuids to detect on.
        callback_url (url) : url that will be called when detection succeeds or fails
    """
    # Check UUIDs
    ibs.web_check_uuids(image_uuid_list=image_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    image_uuid_list = ensure_uuid_list(image_uuid_list)
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    args = (
        gid_list,
        kwargs,
    )
    jobid = ibs.job_manager.jobiface.queue_job(
        'detect_cnn_yolo_json', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/api/engine/labeler/cnn/', methods=['POST'])
def start_labeler_cnn(
    ibs, annot_uuid_list, callback_url=None, callback_method=None, lane='fast', **kwargs
):
    # Check UUIDs
    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    annot_uuid_list = ensure_uuid_list(annot_uuid_list)
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    args = (
        aid_list,
        kwargs,
    )
    jobid = ibs.job_manager.jobiface.queue_job(
        'labeler_cnn', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/api/engine/review/query/chip/best/', methods=['POST'])
def start_review_query_chips_best(
    ibs, annot_uuid, callback_url=None, callback_method=None, lane='slow', **kwargs
):
    annot_uuid_list = [annot_uuid]

    # Check UUIDs
    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    annot_uuid_list = ensure_uuid_list(annot_uuid_list)
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    aid = aid_list[0]
    args = (aid,)
    jobid = ibs.job_manager.jobiface.queue_job(
        'review_query_chips_best', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/test/engine/detect/cnn/yolo/', methods=['GET'])
def start_detect_image_test_yolo(ibs):
    from random import shuffle  # NOQA

    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    gid_list = gid_list[:3]
    image_uuid_list = ibs.get_image_uuids(gid_list)
    jobid = ibs.start_detect_image_yolo(image_uuid_list)
    return jobid


@register_ibs_method
@register_api('/api/engine/detect/cnn/lightnet/', methods=['POST', 'GET'])
def start_detect_image_lightnet(
    ibs, image_uuid_list, callback_url=None, callback_method=None, lane='fast', **kwargs
):
    """
    REST:
        Method: GET/api/engine/detect/cnn/lightnet/
        URL:

    Args:
        image_uuid_list (list) : list of image uuids to detect on.
        callback_url (url) : url that will be called when detection succeeds or fails
    """
    # Check UUIDs
    ibs.web_check_uuids(image_uuid_list=image_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    image_uuid_list = ensure_uuid_list(image_uuid_list)
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    args = (
        gid_list,
        kwargs,
    )
    jobid = ibs.job_manager.jobiface.queue_job(
        'detect_cnn_lightnet_json', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/test/engine/detect/cnn/lightnet/', methods=['GET'])
def start_detect_image_test_lightnet(ibs):
    from random import shuffle  # NOQA

    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    gid_list = gid_list[:3]
    image_uuid_list = ibs.get_image_uuids(gid_list)
    jobid = ibs.start_detect_image_lightnet(image_uuid_list)
    return jobid


@register_ibs_method
@register_api('/api/engine/classify/whaleshark/injury/', methods=['POST'])
def start_predict_ws_injury_interim_svm(
    ibs, annot_uuid_list, callback_url=None, callback_method=None, lane='fast', **kwargs
):
    """
    REST:
        Method: POST
        URL: /api/engine/classify/whaleshark/injury/

    Args:
        annot_uuid_list (list) : list of annot uuids to detect on.
        callback_url (url) : url that will be called when detection succeeds or fails

    CommandLine:
        python -m wbia.web.apis_engine start_predict_ws_injury_interim_svm


    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_engine import *  # NOQA
        >>> from wbia.web import apis_engine
        >>> import wbia
        >>> ibs, qaids, daids = wbia.testdata_expanded_aids(
        >>>     defaultdb='WS_ALL', a=['default:qsize=2,dsize=10'])
        >>> annot_uuid_list = ibs.get_annot_uuids(qaids)
        >>> ibs.initialize_job_manager()
        >>> jobid = ibs.start_predict_ws_injury_interim_svm(annot_uuid_list)
        >>> result = ibs.wait_for_job_result(jobid, timeout=None, freq=2)
        >>> print(result)
        >>> import utool as ut
        >>> #print(ut.to_json(result))
        >>> ibs.close_job_manager()
    """
    # Check UUIDs
    ibs.web_check_uuids(qannot_uuid_list=annot_uuid_list)

    # import wbia
    # from wbia.web import apis_engine
    # ibs.load_plugin_module(apis_engine)
    annot_uuid_list = ensure_uuid_list(annot_uuid_list)
    annots = ibs.annots(uuids=annot_uuid_list)
    args = (annots.aids,)
    jobid = ibs.job_manager.jobiface.queue_job(
        'predict_ws_injury_interim_svm', callback_url, callback_method, lane, *args
    )

    # if callback_url is not None:
    #    #import requests
    #    #requests.
    #    #callback_url
    return jobid


@register_ibs_method
@register_api('/api/engine/query/web/', methods=['GET'])
def start_web_query_all(ibs):
    """
    REST:
        Method: GET
        URL: /api/engine/query/web/
    """
    jobid = ibs.job_manager.jobiface.queue_job('load_identification_query_object_worker')
    return jobid


@register_ibs_method
@register_api('/api/engine/flukebook/sync/', methods=['GET'])
def start_flukebook_sync(ibs, **kwargs):
    """
    REST:
        Method: GET
        URL: /api/engine/flukebook/sync/
    """
    jobid = ibs.job_manager.jobiface.queue_job('flukebook_sync')
    return jobid



# -*- coding: utf-8 -*-
"""
Interface to Darknet object proposals.
"""
import logging
import utool as ut
from os import listdir
from os.path import join, isfile, isdir

(print, rrr, profile) = ut.inject2(__name__, '[rf]')
logger = logging.getLogger('wbia')


VERBOSE_SVM = ut.get_argflag('--verbrf') or ut.VERBOSE


CONFIG_URL_DICT = {
    # 'localizer-zebra-10'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.10.zip',
    # 'localizer-zebra-20'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.20.zip',
    # 'localizer-zebra-30'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.30.zip',
    # 'localizer-zebra-40'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.40.zip',
    # 'localizer-zebra-50'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.50.zip',
    # 'localizer-zebra-60'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.60.zip',
    # 'localizer-zebra-70'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.70.zip',
    # 'localizer-zebra-80'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.80.zip',
    # 'localizer-zebra-90'  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.90.zip',
    # 'localizer-zebra-100' : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.localization.zebra.100.zip',
    # 'image-zebra'         : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.image.zebra.pkl',
    # 'default'             : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.image.zebra.pkl',
    # None                  : 'https://wildbookiarepository.azureedge.net/models/classifier.rf.image.zebra.pkl',
}


def classify_helper(weight_filepath, vector_list, index_list=None, verbose=VERBOSE_SVM):
    if index_list is None:
        index_list = list(range(len(vector_list)))
    # Init score and class holders
    score_dict = {index: [] for index in index_list}
    class_dict = {index: [] for index in index_list}
    # Load models
    model_tup = ut.load_cPkl(weight_filepath, verbose=verbose)
    model, scaler = model_tup
    # Normalize
    vector_list = scaler.transform(vector_list)
    # calculate decisions and predictions
    # score_list = model.decision_function(vector_list)
    score_list = model.predict_proba(vector_list)
    # Take only the positive probability
    score_list = score_list[:, 1]
    class_list = model.predict(vector_list)
    # Zip together results
    zipped = zip(index_list, score_list, class_list)
    for index, score_, class_ in zipped:
        score_dict[index].append(score_)
        class_dict[index].append(class_)
    # Return scores and classes
    return score_dict, class_dict


def classify(vector_list, weight_filepath, verbose=VERBOSE_SVM, **kwargs):
    """
    Args:
        thumbail_list (list of str): the list of image thumbnails that need classifying

    Returns:
        iter
    """
    import multiprocessing
    import numpy as np

    # Get correct weight if specified with shorthand
    if weight_filepath in CONFIG_URL_DICT:
        weight_url = CONFIG_URL_DICT[weight_filepath]
        if weight_url.endswith('.zip'):
            weight_filepath = ut.grab_zipped_url(weight_url, appname='wbia')
        else:
            weight_filepath = ut.grab_file_url(
                weight_url, appname='wbia', check_hash=True
            )

    # Get ensemble
    is_ensemble = isdir(weight_filepath)
    if is_ensemble:
        weight_filepath_list = sorted(
            [
                join(weight_filepath, filename)
                for filename in listdir(weight_filepath)
                if isfile(join(weight_filepath, filename))
            ]
        )
    else:
        weight_filepath_list = [weight_filepath]
    num_weights = len(weight_filepath_list)
    assert num_weights > 0

    # Form dictionaries
    num_vectors = len(vector_list)
    index_list = list(range(num_vectors))

    # Generate parallelized wrapper
    OLD = False
    if is_ensemble and OLD:
        vectors_list = [vector_list for _ in range(num_weights)]
        args_list = zip(weight_filepath_list, vectors_list)
        nTasks = num_weights
        logger.info(
            'Processing ensembles in parallel using %d ensembles' % (num_weights,)
        )
    else:
        num_cpus = multiprocessing.cpu_count()
        vector_batch = int(np.ceil(float(num_vectors) / num_cpus))
        vector_rounds = int(np.ceil(float(num_vectors) / vector_batch))

        args_list = []
        for vector_round in range(vector_rounds):
            start_index = vector_round * vector_batch
            stop_index = (vector_round + 1) * vector_batch
            assert start_index < num_vectors
            stop_index = min(stop_index, num_vectors)
            # logger.info('Slicing index range: [%r, %r)' % (start_index, stop_index, ))

            # Slice gids and get feature data
            index_list_ = list(range(start_index, stop_index))
            vector_list_ = vector_list[start_index:stop_index]
            assert len(index_list_) == len(vector_list_)
            for weight_filepath in weight_filepath_list:
                args = (weight_filepath, vector_list_, index_list_)
                args_list.append(args)

        nTasks = len(args_list)
        logger.info(
            'Processing vectors in parallel using vector_batch = %r' % (vector_batch,)
        )

    # Perform inference
    classify_iter = ut.generate2(
        classify_helper, args_list, nTasks=nTasks, ordered=True, force_serial=True
    )

    # Classify with SVM for each image vector
    score_dict = {index: [] for index in index_list}
    class_dict = {index: [] for index in index_list}
    for score_dict_, class_dict_ in classify_iter:
        for index in index_list:
            if index in score_dict_:
                score_dict[index] += score_dict_[index]
            if index in class_dict_:
                class_dict[index] += class_dict_[index]

    # Organize and compute mode and average for class and score
    for index in index_list:
        score_list_ = score_dict[index]
        class_list_ = class_dict[index]
        score_ = sum(score_list_) / len(score_list_)
        class_ = max(set(class_list_), key=class_list_.count)
        class_ = 'positive' if int(class_) == 1 else 'negative'
        yield score_, class_



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import cv2
import os
import math
import xml.etree.ElementTree as xml

from wbia.detecttools.pascaldata import common as com
from wbia.detecttools.pascaldata.pascal_object import PASCAL_Object


class PASCAL_Image(object):  # NOQA
    def __init__(pascali, filename_xml, absolute_dataset_path, **kwargs):
        with open(filename_xml, 'r') as _xml:
            _xml = xml.XML(_xml.read().replace('\n', ''))

            pascali.folder = com.get(_xml, 'folder')
            pascali.absolute_dataset_path = absolute_dataset_path
            pascali.filename = com.get(_xml, 'filename')

            source = com.get(_xml, 'source', text=False)
            pascali.source_database = com.get(source, 'database')
            pascali.source_annotation = com.get(source, 'annotation')
            pascali.source_image = com.get(source, 'image')

            size = com.get(_xml, 'size', text=False)
            pascali.width = int(com.get(size, 'width'))
            pascali.height = int(com.get(size, 'height'))
            pascali.depth = int(com.get(size, 'depth'))

            pascali.segmented = com.get(size, 'segmented') == '1'

            pascali.objects = []
            pascali.objects_patches = []
            pascali.objects_invalid = []
            for obj in com.get(_xml, 'object', text=False, singularize=False):
                temp = PASCAL_Object(obj, pascali.width, pascali.height)
                if (
                    temp.width > kwargs['object_min_width']
                    and temp.height > kwargs['object_min_height']
                ):
                    pascali.objects.append(temp)
                else:
                    pascali.objects_invalid.append(temp)

            flag = True
            for cat in pascali.categories():
                if cat in kwargs['mine_exclude_categories']:
                    flag = False

            if kwargs['mine_negatives'] and flag:
                negatives = 0
                for i in range(kwargs['mine_max_attempts']):
                    if negatives >= kwargs['mine_max_keep']:
                        break

                    width = com.randInt(
                        kwargs['mine_width_min'],
                        min(pascali.width - 1, kwargs['mine_width_max']),
                    )
                    height = com.randInt(
                        kwargs['mine_height_min'],
                        min(pascali.height - 1, kwargs['mine_height_max']),
                    )
                    x = com.randInt(0, pascali.width - width - 1)
                    y = com.randInt(0, pascali.height - height - 1)

                    obj = {
                        'xmax': x + width,
                        'xmin': x,
                        'ymax': y + height,
                        'ymin': y,
                    }

                    overlap_names = pascali._overlaps(
                        pascali.objects, obj, kwargs['mine_overlap_margin']
                    )
                    if len(overlap_names) > 0:
                        continue

                    pascali.objects.append(
                        PASCAL_Object(obj, pascali.width, pascali.height, name='MINED')
                    )
                    negatives += 1

            if kwargs['mine_patches']:
                patch_width = kwargs['mine_patch_width']
                patch_height = kwargs['mine_patch_height']
                x_length = float(pascali.width - patch_width - 1)
                y_length = float(pascali.height - patch_height - 1)
                x_bins = int(x_length / kwargs['mine_patch_stride_suggested'])
                y_bins = int(y_length / kwargs['mine_patch_stride_suggested'])

                x_bins = max(1, x_bins)
                y_bins = max(1, y_bins)
                patch_stride_x = x_length / x_bins
                patch_stride_y = y_length / y_bins
                # pascali.show()
                for x in range(x_bins + 1):
                    for y in range(y_bins + 1):
                        x_min = int(x * patch_stride_x)
                        y_min = int(y * patch_stride_y)
                        x_max = x_min + patch_width
                        y_max = y_min + patch_height
                        assert (
                            0 <= x_min
                            and x_max < pascali.width
                            and 0 <= y_min
                            and y_max < pascali.height
                        )
                        # Add patch
                        obj = {
                            'xmax': x_max,
                            'xmin': x_min,
                            'ymax': y_max,
                            'ymin': y_min,
                        }
                        overlap_names = pascali._overlaps(
                            pascali.objects, obj, kwargs['mine_patch_overlap_margin']
                        )
                        if len(overlap_names) > 0:
                            for overlap_name in overlap_names:
                                name = '%s' % overlap_name.upper()
                                pascali.objects_patches.append(
                                    PASCAL_Object(
                                        obj, pascali.width, pascali.height, name=name
                                    )
                                )
                        else:
                            pascali.objects_patches.append(
                                PASCAL_Object(
                                    obj, pascali.width, pascali.height, name='NEGATIVE'
                                )
                            )

    def __str__(pascali):
        return '<IBEIS Image Object | %s | %d objects>' % (
            pascali.filename,
            len(pascali.objects),
        )

    def __repr__(pascali):
        return '<IBEIS Image Object | %s>' % (pascali.filename)

    def __len__(pascali):
        return len(pascali.objects)

    def _distance(p1, p2):
        x1, y1 = p1
        x2, y2 = p2
        return math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    def _overlaps(
        pascali, objects, obj, margin=0.50, bins=['left', 'front', 'right', 'back']
    ):
        names = []
        for _obj in objects:
            x_overlap = max(0, min(obj['xmax'], _obj.xmax) - max(obj['xmin'], _obj.xmin))
            y_overlap = max(0, min(obj['ymax'], _obj.ymax) - max(obj['ymin'], _obj.ymin))
            area_overlap = float(x_overlap * y_overlap)
            width = obj['xmax'] - obj['xmin']
            height = obj['ymax'] - obj['ymin']
            area_total = min(width * height, _obj.area)
            score = area_overlap / area_total
            # print(score)
            if score >= margin:
                names.append(_obj.name + ':' + _obj.pose_str)
        return list(set(names))

    def image_path(pascali):
        return os.path.join(pascali.absolute_dataset_path, 'JPEGImages', pascali.filename)

    def categories(pascali, unique=True, patches=False):
        temp = [_object.name for _object in pascali.objects]
        if patches:
            temp += [_object.name for _object in pascali.objects_patches]
        if unique:
            temp = set(temp)
        return sorted(temp)

    def bounding_boxes(pascali, parts=False):
        return [_object.bounding_box(parts) for _object in pascali.objects]

    def _accuracy_match(pascali, prediction, object_list):

        # For this non-supressed prediction, compute and assign to the closest bndbox
        centerx, centery, minx, miny, maxx, maxy, confidence, supressed = prediction

        index_best = None
        score_best = -1.0
        for index, _object in enumerate(object_list):
            width = maxx - minx
            height = maxy - miny

            x_overlap = max(0, min(maxx, _object.xmax) - max(minx, _object.xmin))
            y_overlap = max(0, min(maxy, _object.ymax) - max(miny, _object.ymin))
            area_overlap = float(x_overlap * y_overlap)
            area_total = (width * height) + _object.area
            score = area_overlap / (area_total - area_overlap)

            if score >= score_best:
                # Wooo! Found a (probably) better candidate, but...
                if score == score_best:
                    # Well, this is awkward?
                    assert index_best is not None  # Just to be sure
                    _object_best = object_list[index_best]

                    a = pascali._distance(
                        (centerx, centery), (_object_best.xcenter, _object_best.ycenter)
                    )
                    b = pascali._distance(
                        (centerx, centery), (_object.xcenter, _object.ycenter)
                    )
                    if a < b:
                        # Not a better candidate based on distance
                        continue
                    elif a == b:
                        # First come, first serve
                        continue
                # Save new best
                score_best = score
                index_best = index

        return index_best, score_best

    def accuracy(pascali, prediction_list, category, alpha=0.5):
        # PASCAL ACCURACY MEASUREMENT
        object_list = []
        for _object in pascali.objects + pascali.objects_invalid:
            if _object.name == category:
                object_list.append(_object)

        # Trivial case
        if len(object_list) == 0 and len(prediction_list) == 0:
            return 1.0, 0.0, 0.0, 0.0

        true_positive = 0
        false_positive = 0

        counters = [0] * len(object_list)
        for prediction in prediction_list:
            centerx, centery, minx, miny, maxx, maxy, confidence, supressed = prediction
            if supressed == 0.0:
                index_best, score_best = pascali._accuracy_match(prediction, object_list)
                if score_best >= alpha:
                    counters[index_best] += 1
                    true_positive += 1
                else:
                    false_positive += 1

        false_negative = counters.count(0)
        precision = float(true_positive)
        recall = true_positive + false_positive + false_negative
        assert recall != 0
        return precision / recall, true_positive, false_positive, false_negative

    def show(
        pascali,
        objects=True,
        parts=True,
        display=True,
        prediction_list=None,
        category=None,
        alpha=0.5,
    ):
        def _draw_box(img, annotation, xmin, ymin, xmax, ymax, color, stroke=2, top=True):
            font = cv2.FONT_HERSHEY_SIMPLEX
            scale = 0.5
            width, height = cv2.getTextSize(annotation, font, scale, -1)[0]
            cv2.rectangle(img, (xmin, ymin), (xmax, ymax), color, stroke)
            if top:
                cv2.rectangle(img, (xmin, ymin - height), (xmin + width, ymin), color, -1)
                cv2.putText(img, annotation, (xmin + 5, ymin), font, 0.4, (255, 255, 255))
            else:
                cv2.rectangle(img, (xmin, ymax - height), (xmin + width, ymax), color, -1)
                cv2.putText(img, annotation, (xmin + 5, ymax), font, 0.4, (255, 255, 255))

        original = com.openImage(pascali.image_path(), color=True)
        color_dict = {}
        for _object in pascali.objects:
            color = com.randColor()
            color_dict[_object] = color
            _draw_box(
                original,
                _object.name.upper(),
                _object.xmin,
                _object.ymin,
                _object.xmax,
                _object.ymax,
                color,
            )

            if parts:
                for part in _object.parts:
                    _draw_box(
                        original,
                        part.name.upper(),
                        part.xmin,
                        part.ymin,
                        part.xmax,
                        part.ymax,
                        color,
                    )

        for _object in pascali.objects_invalid:
            color = [0, 0, 0]
            color_dict[_object] = color
            _draw_box(
                original,
                _object.name.upper(),
                _object.xmin,
                _object.ymin,
                _object.xmax,
                _object.ymax,
                color,
            )

            if parts:
                for part in _object.parts:
                    _draw_box(
                        original,
                        part.name.upper(),
                        part.xmin,
                        part.ymin,
                        part.xmax,
                        part.ymax,
                        color,
                    )

        for _object in pascali.objects_patches:
            if _object.name.upper() == 'NEGATIVE':
                continue
                color = [255, 0, 0]
            else:
                color = [0, 0, 255]
            color_dict[_object] = color
            _draw_box(
                original,
                _object.name.upper(),
                _object.xmin,
                _object.ymin,
                _object.xmax,
                _object.ymax,
                color,
            )

            if parts:
                for part in _object.parts:
                    _draw_box(
                        original,
                        part.name.upper(),
                        part.xmin,
                        part.ymin,
                        part.xmax,
                        part.ymax,
                        color,
                    )

        if prediction_list is not None:
            assert category is not None
            object_list = []
            for _object in pascali.objects + pascali.objects_invalid:
                if _object.name == category:
                    object_list.append(_object)

            for prediction in prediction_list:
                (
                    centerx,
                    centery,
                    minx,
                    miny,
                    maxx,
                    maxy,
                    confidence,
                    supressed,
                ) = prediction
                if supressed == 0.0:
                    if len(object_list) > 0:
                        index_best, score_best = pascali._accuracy_match(
                            prediction, object_list
                        )
                        _object_best = object_list[index_best]
                        color = color_dict[_object_best]
                        if score_best >= alpha:
                            annotation = 'DETECT [TRUE POS %.2f]' % score_best
                        else:
                            annotation = 'DETECT [FALSE POS %.2f]' % score_best
                        cv2.line(
                            original,
                            (int(minx), int(miny)),
                            (_object_best.xmin, _object_best.ymin),
                            color,
                            1,
                        )
                        cv2.line(
                            original,
                            (int(minx), int(maxy)),
                            (_object_best.xmin, _object_best.ymax),
                            color,
                            1,
                        )
                        cv2.line(
                            original,
                            (int(maxx), int(miny)),
                            (_object_best.xmax, _object_best.ymin),
                            color,
                            1,
                        )
                        cv2.line(
                            original,
                            (int(maxx), int(maxy)),
                            (_object_best.xmax, _object_best.ymax),
                            color,
                            1,
                        )

                    else:
                        annotation = 'DETECT [FALSE POS]'
                        color = [0, 0, 255]
                    _draw_box(
                        original,
                        annotation,
                        int(minx),
                        int(miny),
                        int(maxx),
                        int(maxy),
                        color,
                        stroke=1,
                        top=False,
                    )

        if display:
            cv2.imshow(pascali.filename + ' with Bounding Boxes', original)
            cont = input()
            cv2.destroyAllWindows()
            return cont == ''
        else:
            return original



# -*- coding: utf-8 -*-
"""
"""
import logging
import utool as ut

(print, rrr, profile) = ut.inject2(__name__, '[preproc_residual]')
logger = logging.getLogger('wbia')


def add_residual_params_gen(ibs, fid_list, qreq_=None):
    return None


def on_delete(ibs, featweight_rowid_list):
    logger.info('Warning: Not Implemented')
    logger.info('Probably nothing to do here')



# -*- coding: utf-8 -*-
"""
old code ported from utool
"""
import sys
import six
import traceback
from utool.Preferences import Pref, PrefNode, PrefChoice
from wbia.guitool.__PYQT__ import _fromUtf8, _translate, QVariantHack
from wbia.guitool.__PYQT__ import QtWidgets, QtCore
from wbia.guitool.__PYQT__.QtCore import Qt
from wbia.guitool import qtype
import utool as ut
from utool import util_type

ut.noinject(__name__, '[PreferenceWidget]', DEBUG=False)

VERBOSE_PREF = ut.get_argflag('--verbpref')


def report_thread_error(fn):
    """ Decorator to help catch errors that QT wont report """

    def report_thread_error_wrapper(*args, **kwargs):
        try:
            ret = fn(*args, **kwargs)
            return ret
        except Exception as ex:
            print('\n\n *!!* Thread Raised Exception: ' + str(ex))
            print('\n\n *!!* Thread Exception Traceback: \n\n' + traceback.format_exc())
            sys.stdout.flush()
            et, ei, tb = sys.exc_info()
            raise

    return report_thread_error_wrapper


def _qt_set_leaf_data(self, qvar):
    """ Sets backend data using QVariants """
    if VERBOSE_PREF:
        print('')
        print('+--- [pref.qt_set_leaf_data]')
        print('[pref.qt_set_leaf_data] qvar = %r' % qvar)
        print('[pref.qt_set_leaf_data] _intern.name=%r' % self._intern.name)
        print('[pref.qt_set_leaf_data] _intern.type_=%r' % self._intern.get_type())
        print('[pref.qt_set_leaf_data] type(_intern.value)=%r' % type(self._intern.value))
        print('[pref.qt_set_leaf_data] _intern.value=%r' % self._intern.value)
        # print('[pref.qt_set_leaf_data] qvar.toString()=%s' % str(qvar.toString()))
    if self._tree.parent is None:
        raise Exception('[Pref.qtleaf] Cannot set root preference')
    if self.qt_is_editable():
        new_val = '[Pref.qtleaf] BadThingsHappenedInPref'
        if self._intern.value == PrefNode:
            raise Exception('[Pref.qtleaf] Qt can only change leafs')
        elif self._intern.value is None:
            # None could be a number of types
            def cast_order(var, order=[bool, int, float, str]):
                for type_ in order:
                    try:
                        ret = type_(var)
                        return ret
                    except Exception:
                        continue

            new_val = cast_order(str(qvar))
        self._intern.get_type()
        if isinstance(self._intern.value, bool):
            # new_val = bool(qvar.toBool())
            print('qvar = %r' % (qvar,))
            new_val = util_type.smart_cast(qvar, bool)
            # new_val = bool(eval(qvar, {}, {}))
            print('new_val = %r' % (new_val,))
        elif isinstance(self._intern.value, int):
            # new_val = int(qvar.toInt()[0])
            new_val = int(qvar)
        # elif isinstance(self._intern.value, float):
        elif self._intern.get_type() in util_type.VALID_FLOAT_TYPES:
            # new_val = float(qvar.toDouble()[0])
            new_val = float(qvar)
        elif isinstance(self._intern.value, six.string_types):
            # new_val = str(qvar.toString())
            new_val = str(qvar)
        elif isinstance(self._intern.value, PrefChoice):
            # new_val = qvar.toString()
            new_val = str(qvar)
            if new_val.upper() == 'NONE':
                new_val = None
        else:
            try:
                # new_val = str(qvar.toString())
                type_ = self._intern.get_type()
                if type_ is not None:
                    new_val = type_(str(qvar))
                else:
                    new_val = str(qvar)
            except Exception:
                raise NotImplementedError(
                    (
                        '[Pref.qtleaf] Unknown internal type. '
                        'type(_intern.value) = %r, '
                        '_intern.get_type() = %r, '
                    )
                    % type(self._intern.value),
                    self._intern.get_type(),
                )
        # Check for a set of None
        if isinstance(new_val, six.string_types):
            if new_val.lower() == 'none':
                new_val = None
            elif new_val.lower() == 'true':
                new_val = True
            elif new_val.lower() == 'false':
                new_val = False
        # save to disk after modifying data
        if VERBOSE_PREF:
            print('---')
            print('[pref.qt_set_leaf_data] new_val=%r' % new_val)
            print('[pref.qt_set_leaf_data] type(new_val)=%r' % type(new_val))
            print('L____ [pref.qt_set_leaf_data]')
        # TODO Add ability to set a callback function when certain
        # preferences are changed.
        return self._tree.parent.pref_update(self._intern.name, new_val)
    return 'PrefNotEditable'


class QPreferenceModel(QtCore.QAbstractItemModel):
    """ Convention states only items with column index 0 can have children """

    @report_thread_error
    def __init__(self, pref_struct, parent=None):
        super(QPreferenceModel, self).__init__(parent)
        self.rootPref = pref_struct

    @report_thread_error
    def index2Pref(self, index=QtCore.QModelIndex()):
        """ Internal helper method """
        if index.isValid():
            item = index.internalPointer()
            if item:
                return item
        return self.rootPref

    # -----------
    # Overloaded ItemModel Read Functions
    @report_thread_error
    def rowCount(self, parent=QtCore.QModelIndex()):
        parentPref = self.index2Pref(parent)
        return parentPref.qt_row_count()

    @report_thread_error
    def columnCount(self, parent=QtCore.QModelIndex()):
        parentPref = self.index2Pref(parent)
        return parentPref.qt_col_count()

    @report_thread_error
    def data(self, qtindex, role=Qt.DisplayRole):
        """Returns the data stored under the given role
        for the item referred to by the qtindex."""
        if not qtindex.isValid():
            return QVariantHack()
        # Specify CheckState Role:
        flags = self.flags(qtindex)
        if role == Qt.CheckStateRole:
            if flags & Qt.ItemIsUserCheckable:
                data = self.index2Pref(qtindex).qt_get_data(qtindex.column())
                return Qt.Checked if data else Qt.Unchecked
        if role != Qt.DisplayRole and role != Qt.EditRole:
            return QVariantHack()
        nodePref = self.index2Pref(qtindex)
        data = nodePref.qt_get_data(qtindex.column())
        # print('--- data() ---')
        # print('role = %r' % role)
        # print('data = %r' % data)
        # print('type(data) = %r' % type(data))
        # <SIP.API_MODE(1)>
        # var = QVariantHack(data)
        # if isinstance(data, float):
        #    var = QVariantHack(QString.number(data, format='g', precision=6))
        # if isinstance(data, bool):
        #    var = QVariantHack(data).toString()
        # if isinstance(data, int):
        #    var = QVariantHack(data).toString()
        # </SIP.API_MODE(1)>
        # <SIP.API_MODE(2)>
        if isinstance(data, float):
            var = qtype.locale_float(data, 6)
        else:
            var = data
        # </SIP.API_MODE(2)>
        # print('var= %r' % var)
        # print('type(var)= %r' % type(var))
        return str(var)

    @report_thread_error
    def setData(self, qtindex, value, role=Qt.EditRole):
        """Sets the role data for the item at qtindex to value."""
        if role == Qt.EditRole:
            data = value
        elif role == Qt.CheckStateRole:
            data = value == Qt.Checked
        else:
            return False

        if VERBOSE_PREF:
            print('[qt] --- setData() ---')
            print('[qt] role = %r' % role)
            print('[qt] value = %r' % value)
            print('[qt] type(data) = %r' % type(data))
            print('[qt] type(value) = %r' % type(value))

        leafPref = self.index2Pref(qtindex)
        old_data = leafPref.qt_get_data(qtindex.column())
        if VERBOSE_PREF:
            print('[qt] old_data = %r' % (old_data,))
            print('[qt] old_data != data = %r' % (old_data != data,))
        if old_data != data:
            result = _qt_set_leaf_data(leafPref, data)
            if VERBOSE_PREF:
                print('[qt] result = %r' % (result,))
            # if result is True:
            # return result
        self.dataChanged.emit(qtindex, qtindex)
        return True

    @report_thread_error
    def index(self, row, col, parent=QtCore.QModelIndex()):
        """Returns the index of the item in the model specified
        by the given row, column and parent index."""
        if parent.isValid() and parent.column() != 0:
            return QtCore.QModelIndex()
        parentPref = self.index2Pref(parent)
        childPref = parentPref.qt_get_child(row)
        if childPref:
            return self.createIndex(row, col, childPref)
        else:
            return QtCore.QModelIndex()

    @report_thread_error
    def parent(self, index=None):
        """Returns the parent of the model item with the given index.
        If the item has no parent, an invalid QModelIndex is returned."""
        if index is None:  # Overload with QtCore.QObject.parent()
            return QtCore.QObject.parent(self)
        if not index.isValid():
            return QtCore.QModelIndex()
        nodePref = self.index2Pref(index)
        parentPref = nodePref.qt_get_parent()
        if parentPref == self.rootPref:
            return QtCore.QModelIndex()
        return self.createIndex(parentPref.qt_parents_index_of_me(), 0, parentPref)

    @report_thread_error
    def flags(self, index):
        """Returns the item flags for the given index."""
        if index.column() == 0:
            # The First Column is just a label and unchangable
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable
        if not index.isValid():
            return Qt.ItemFlag(0)
        childPref = self.index2Pref(index)
        if childPref:
            if childPref.qt_is_editable():
                if childPref._intern.get_type() is bool:
                    # print(childPref)
                    # print(childPref._intern.get_type())
                    # print(childPref._intern.get_type() is bool)
                    return Qt.ItemIsEnabled | Qt.ItemIsUserCheckable
                else:
                    return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable
        return Qt.ItemFlag(0)

    @report_thread_error
    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section == 0:
                return QVariantHack('Config Key')
            if section == 1:
                return QVariantHack('Config Value')
        return QVariantHack()


class Ui_editPrefSkel(object):
    """
    THE PREFERENCE SKELETON
    """

    def setupUi(self, editPrefSkel):
        editPrefSkel.setObjectName(_fromUtf8('editPrefSkel'))
        editPrefSkel.resize(668, 530)
        # Add Pane for TreeView
        self.verticalLayout = QtWidgets.QVBoxLayout(editPrefSkel)
        self.verticalLayout.setObjectName(_fromUtf8('verticalLayout'))
        # The TreeView for QtCore.QAbstractItemModel to attach to
        self.prefTreeView = QtWidgets.QTreeView(editPrefSkel)
        self.prefTreeView.setObjectName(_fromUtf8('prefTreeView'))
        self.verticalLayout.addWidget(self.prefTreeView)
        # Add Pane for buttons
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8('horizontalLayout'))
        #
        # self.redrawBUT = QtWidgets.QPushButton(editPrefSkel)
        # self.redrawBUT.setObjectName(_fromUtf8('redrawBUT'))
        # self.horizontalLayout.addWidget(self.redrawBUT)
        ##
        # self.unloadFeaturesAndModelsBUT = QtWidgets.QPushButton(editPrefSkel)
        # self.unloadFeaturesAndModelsBUT.setObjectName(_fromUtf8('unloadFeaturesAndModelsBUT'))
        # self.horizontalLayout.addWidget(self.unloadFeaturesAndModelsBUT)
        #
        self.defaultPrefsBUT = QtWidgets.QPushButton(editPrefSkel)
        self.defaultPrefsBUT.setObjectName(_fromUtf8('defaultPrefsBUT'))
        self.horizontalLayout.addWidget(self.defaultPrefsBUT)
        # Buttons are a child of the View
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.retranslateUi(editPrefSkel)
        QtCore.QMetaObject.connectSlotsByName(editPrefSkel)

    def retranslateUi(self, editPrefSkel):
        # UTF-8 Support
        editPrefSkel.setWindowTitle(_translate('editPrefSkel', 'Edit Preferences', None))
        # self.redrawBUT.setText(_translate('editPrefSkel', 'Redraw', None))
        # self.unloadFeaturesAndModelsBUT.setText(_translate('editPrefSkel', 'Unload Features and Models', None))
        self.defaultPrefsBUT.setText(_translate('editPrefSkel', 'Defaults', None))


# ---
# THE PREFERENCE WIDGET
# ---
class EditPrefWidget(QtWidgets.QWidget):
    """The Settings Pane; Subclass of Main Windows."""

    def __init__(self, pref_struct):
        super(EditPrefWidget, self).__init__()
        self.ui = Ui_editPrefSkel()
        self.ui.setupUi(self)
        self.pref_model = None
        self.populatePrefTreeSlot(pref_struct)
        # self.ui.redrawBUT.clicked.connect(fac.redraw)
        # self.ui.defaultPrefsBUT.clicked.connect(fac.default_prefs)
        # self.ui.unloadFeaturesAndModelsBUT.clicked.connect(fac.unload_features_and_models)

    @QtCore.pyqtSlot(Pref, name='populatePrefTreeSlot')
    def populatePrefTreeSlot(self, pref_struct):
        """Populates the Preference Tree Model"""
        # printDBG('Bulding Preference Model of: ' + repr(pref_struct))
        # Creates a QStandardItemModel that you can connect to a QTreeView
        self.pref_model = QPreferenceModel(pref_struct)
        # printDBG('Built: ' + repr(self.pref_model))
        self.ui.prefTreeView.setModel(self.pref_model)
        self.ui.prefTreeView.header().resizeSection(0, 250)

    def refresh_layout(self):
        self.pref_model.layoutAboutToBeChanged.emit()
        self.pref_model.layoutChanged.emit()


def test_preference_gui():
    r"""
    CommandLine:
        python -m wbia.guitool.PreferenceWidget --exec-test_preference_gui --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.guitool.PreferenceWidget import *  # NOQA
        >>> import wbia.guitool
        >>> guitool.ensure_qtapp()
        >>> scope = test_preference_gui()
        >>> ut.quit_if_noshow()
        >>> guitool.qtapp_loop(freq=10)
    """
    from wbia import dtool

    class DtoolConfig(dtool.Config):
        _param_info_list = [
            ut.ParamInfo('str_option2', 'hello'),
            ut.ParamInfo('int_option2', 456),
            ut.ParamInfo('bool_option2', True),
        ]

    class OldPrefConfig(ut.Pref):
        def __init__(self):
            super(OldPrefConfig, self).__init__()
            self.str_option = 'goodbye'
            self.int_option = 123
            self.bool_option = False
            self.float_option = 3.2
            self.listvar = ['one', 'two', 'three']

    old = OldPrefConfig()
    new = DtoolConfig()
    new_wrap = ut.Pref()
    for k, v in new.items():
        setattr(new_wrap, k, v)
    old.new = new_wrap
    epw = old.createQWidget()
    from wbia.plottool import fig_presenter

    fig_presenter.register_qt4_win(epw)
    # epw.ui.defaultPrefsBUT.clicked.connect(back.default_config)
    epw.show()
    return old, epw



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Exports subset of an IBEIS database to a new IBEIS database
"""
import logging
import utool as ut
from wbia.other import ibsfuncs
from wbia import constants as const

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def check_merge(ibs_src, ibs_dst):
    aid_list1 = ibs_src.get_valid_aids()
    gid_list1 = ibs_src.get_annot_gids(aid_list1)
    gname_list1 = ibs_src.get_image_uris(gid_list1)
    image_uuid_list1 = ibs_src.get_image_uuids(gid_list1)
    gid_list2 = ibs_dst.get_image_gids_from_uuid(image_uuid_list1)
    gname_list2 = ibs_dst.get_image_uris(gid_list2)
    # Asserts
    ut.assert_all_not_None(gid_list1, 'gid_list1')
    ut.assert_all_not_None(gid_list2, 'gid_list2')
    ut.assert_lists_eq(gname_list1, gname_list2, 'faild gname')
    # Image UUIDS should be consistent between databases
    image_uuid_list2 = ibs_dst.get_image_uuids(gid_list2)
    ut.assert_lists_eq(image_uuid_list1, image_uuid_list2, 'failed uuid')

    aids_list1 = ibs_src.get_image_aids(gid_list1)
    aids_list2 = ibs_dst.get_image_aids(gid_list2)

    avuuids_list1 = ibs_src.unflat_map(ibs_src.get_annot_visual_uuids, aids_list1)
    avuuids_list2 = ibs_dst.unflat_map(ibs_dst.get_annot_visual_uuids, aids_list2)

    issubset_list = [
        set(avuuids1).issubset(set(avuuids2))
        for avuuids1, avuuids2 in zip(avuuids_list1, avuuids_list2)
    ]
    assert all(issubset_list), 'ibs_src must be a subset of ibs_dst: issubset_list=%r' % (
        issubset_list,
    )
    # aids_depth1 = ut.depth_profile(aids_list1)
    # aids_depth2 = ut.depth_profile(aids_list2)
    # depth might not be true if ibs_dst is not empty
    # ut.assert_lists_eq(aids_depth1, aids_depth2, 'failed depth')
    logger.info('Merge seems ok...')


def merge_databases(ibs_src, ibs_dst, rowid_subsets=None, localize_images=True):
    """
    New way of merging using the non-hacky sql table merge.
    However, its only workings due to major hacks.

    FIXME: annotmatch table

    CommandLine:
        python -m wbia --test-merge_databases

        python -m wbia merge_databases:0 --db1 LF_OPTIMIZADAS_NI_V_E --db2 LF_ALL
        python -m wbia merge_databases:0 --db1 LF_WEST_POINT_OPTIMIZADAS --db2 LF_ALL

        python -m wbia merge_databases:0 --db1 PZ_Master0 --db2 PZ_Master1
        python -m wbia merge_databases:0 --db1 NNP_Master3 --db2 PZ_Master1

        python -m wbia merge_databases:0 --db1 GZ_ALL --db2 GZ_Master1
        python -m wbia merge_databases:0 --db1 lewa_grevys --db2 GZ_Master1

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.dbio.export_subset import *  # NOQA
        >>> import wbia
        >>> db1 = ut.get_argval('--db1', str, default=None)
        >>> db2 = ut.get_argval('--db2', str, default=None)
        >>> dbdir1 = ut.get_argval('--dbdir1', str, default=None)
        >>> dbdir2 = ut.get_argval('--dbdir2', str, default=None)
        >>> delete_ibsdir = False
        >>> # Check for test mode instead of script mode
        >>> if db1 is None and db2 is None and dbdir1 is None and dbdir2 is None:
        ...     db1 = 'testdb1'
        ...     dbdir2 = 'testdb_dst'
        ...     delete_ibsdir = True
        >>> # Open the source and destination database
        >>> assert db1 is not None or dbdir1 is not None
        >>> assert db2 is not None or dbdir2 is not None
        >>> ibs_src = wbia.opendb(db=db1, dbdir=dbdir1)
        >>> ibs_dst = wbia.opendb(db=db2, dbdir=dbdir2, allow_newdir=True,
        >>>                        delete_ibsdir=delete_ibsdir)
        >>> merge_databases(ibs_src, ibs_dst)
        >>> check_merge(ibs_src, ibs_dst)
        >>> ibs_dst.print_dbinfo()
    """
    # TODO: ensure images are localized
    # otherwise this wont work
    logger.info(
        'BEGIN MERGE OF %r into %r' % (ibs_src.get_dbname(), ibs_dst.get_dbname())
    )
    # ibs_src.run_integrity_checks()
    # ibs_dst.run_integrity_checks()
    ibs_dst.update_annot_visual_uuids(ibs_dst.get_valid_aids())
    ibs_src.update_annot_visual_uuids(ibs_src.get_valid_aids())
    ibs_src.ensure_contributor_rowids()
    ibs_dst.ensure_contributor_rowids()
    ibs_src.fix_invalid_annotmatches()
    ibs_dst.fix_invalid_annotmatches()
    # Hack move of the external data
    if rowid_subsets is not None and const.IMAGE_TABLE in rowid_subsets:
        gid_list = rowid_subsets[const.IMAGE_TABLE]
    else:
        gid_list = ibs_src.get_valid_gids()
    imgpath_list = ibs_src.get_image_paths(gid_list)
    dst_imgdir = ibs_dst.get_imgdir()
    if localize_images:
        ut.copy_files_to(imgpath_list, dst_imgdir, overwrite=False, verbose=True)
    ignore_tables = [
        'lblannot',
        'lblimage',
        'image_lblimage_relationship',
        'annotation_lblannot_relationship',
        'keys',
    ]
    # ignore_tables += [
    #     'contributors', 'party', 'configs'
    # ]
    # TODO: Fix database merge to allow merging tables with more than one superkey
    # and no primary superkey.
    error_tables = [
        'imageset_image_relationship',
        'annotgroup_annotation_relationship',
        'annotmatch',
    ]
    ignore_tables += error_tables
    ibs_dst.db.merge_databases_new(
        ibs_src.db, ignore_tables=ignore_tables, rowid_subsets=rowid_subsets
    )
    logger.info(
        'FINISHED MERGE %r into %r' % (ibs_src.get_dbname(), ibs_dst.get_dbname())
    )


def make_new_dbpath(ibs, id_label, id_list):
    """
    Creates a new database path unique to the exported subset of ids.
    """
    import wbia

    tag_hash = ut.hashstr_arr(id_list, hashlen=8, alphabet=ut.ALPHABET_27)
    base_fmtstr = (
        ibs.get_dbname()
        + '_'
        + id_label
        + 's='
        + tag_hash.replace('(', '_').replace(')', '_')
        + '_%d'
    )
    dpath = wbia.get_workdir()
    new_dbpath = ut.non_existing_path(base_fmtstr, dpath)
    return new_dbpath


def export_names(ibs, nid_list, new_dbpath=None):
    r"""
    exports a subset of names and other required info

    Args:
        ibs (IBEISController):  wbia controller object
        nid_list (list):

    CommandLine:
        python -m wbia.dbio.export_subset --test-export_names

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.dbio.export_subset import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb2')
        >>> ibs.delete_empty_nids()
        >>> nid_list = ibs._get_all_known_nids()[0:2]
        >>> # execute function
        >>> result = export_names(ibs, nid_list)
        >>> # verify results
        >>> print(result)
    """
    logger.info('Exporting name nid_list=%r' % (nid_list,))
    if new_dbpath is None:
        new_dbpath = make_new_dbpath(ibs, 'nid', nid_list)

    aid_list = ut.flatten(ibs.get_name_aids(nid_list))
    gid_list = ut.unique_unordered(ibs.get_annot_gids(aid_list))

    return export_data(ibs, gid_list, aid_list, nid_list, new_dbpath=new_dbpath)


def find_gid_list(ibs, min_count=500, ensure_annots=False):
    import random

    gid_list = ibs.get_valid_gids()
    reviewed_list = ibs.get_image_reviewed(gid_list)

    if ensure_annots:
        aids_list = ibs.get_image_aids(gid_list)
        reviewed_list = [
            0 if len(aids) == 0 else reviewed
            for aids, reviewed in zip(aids_list, reviewed_list)
        ]

    # Filter by reviewed
    gid_list = [gid for gid, reviewed in zip(gid_list, reviewed_list) if reviewed == 1]

    if len(gid_list) < min_count:
        return None

    while len(gid_list) > min_count:
        index = random.randint(0, len(gid_list) - 1)
        del gid_list[index]

    return gid_list


def __export_reviewed_subset(ibs, min_count=500, ensure_annots=False):
    from os.path import join

    gid_list = find_gid_list(ibs, min_count=min_count, ensure_annots=ensure_annots)
    if gid_list is None:
        return None
    new_dbpath = '/' + join('Datasets', 'BACKGROUND', ibs.dbname)
    logger.info('Exporting to %r with %r images' % (new_dbpath, len(gid_list)))
    return export_images(ibs, gid_list, new_dbpath=new_dbpath)


def export_images(ibs, gid_list, new_dbpath=None):
    """
    exports a subset of images and other required info

    TODO:
        PZ_Master1 needs to backproject information back on to NNP_Master3 and PZ_Master0

    Args:
        ibs (IBEISController):  wbia controller object
        gid_list (list):  list of annotation rowids
        new_dbpath (None): (default = None)

    Returns:
        str: new_dbpath
    """
    logger.info('Exporting image gid_list=%r' % (gid_list,))
    if new_dbpath is None:
        new_dbpath = make_new_dbpath(ibs, 'gid', gid_list)

    aid_list = ut.unique_unordered(ut.flatten(ibs.get_image_aids(gid_list)))
    nid_list = ut.unique_unordered(ibs.get_annot_nids(aid_list))

    return export_data(ibs, gid_list, aid_list, nid_list, new_dbpath=new_dbpath)


def export_annots(ibs, aid_list, new_dbpath=None):
    r"""
    exports a subset of annotations and other required info

    TODO:
        PZ_Master1 needs to backproject information back on to NNP_Master3 and
        PZ_Master0

    Args:
        ibs (IBEISController):  wbia controller object
        aid_list (list):  list of annotation rowids
        new_dbpath (None): (default = None)

    Returns:
        str: new_dbpath

    CommandLine:
        python -m wbia.dbio.export_subset export_annots
        python -m wbia.dbio.export_subset export_annots --db NNP_Master3 \
            -a viewpoint_compare --nocache-aid --verbtd --new_dbpath=PZ_ViewPoints

        python -m wbia.expt.experiment_helpers get_annotcfg_list:0 \
            --db NNP_Master3 \
            -a viewpoint_compare --nocache-aid --verbtd
        python -m wbia.expt.experiment_helpers get_annotcfg_list:0 --db NNP_Master3 \
            -a viewpoint_compare --nocache-aid --verbtd
        python -m wbia.expt.experiment_helpers get_annotcfg_list:0 --db NNP_Master3 \
            -a default:aids=all,is_known=True,view_pername=#primary>0&#primary1>0,per_name=4,size=200
        python -m wbia.expt.experiment_helpers get_annotcfg_list:0 --db NNP_Master3 \
            -a default:aids=all,is_known=True,view_pername='#primary>0&#primary1>0',per_name=4,size=200 --acfginfo

        python -m wbia.expt.experiment_helpers get_annotcfg_list:0 --db PZ_Master1 \
            -a default:has_any=photobomb --acfginfo

    Example:
        >>> # SCRIPT
        >>> from wbia.dbio.export_subset import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import experiment_helpers
        >>> ibs = wbia.opendb(defaultdb='NNP_Master3')
        >>> acfg_name_list = ut.get_argval(('--aidcfg', '--acfg', '-a'), type_=list, default=[''])
        >>> acfg_list, expanded_aids_list = experiment_helpers.get_annotcfg_list(ibs, acfg_name_list)
        >>> aid_list = expanded_aids_list[0][0]
        >>> ibs.print_annot_stats(aid_list, viewcode_isect=True, per_image=True)
        >>> # Expand to get all annots in each chosen image
        >>> gid_list = ut.unique_ordered(ibs.get_annot_gids(aid_list))
        >>> aid_list = ut.flatten(ibs.get_image_aids(gid_list))
        >>> ibs.print_annot_stats(aid_list, viewcode_isect=True, per_image=True)
        >>> new_dbpath = ut.get_argval('--new-dbpath', default='PZ_ViewPoints')
        >>> new_dbpath = export_annots(ibs, aid_list, new_dbpath)
        >>> result = ('new_dbpath = %s' % (str(new_dbpath),))
        >>> print(result)
    """
    logger.info('Exporting annotations aid_list=%r' % (aid_list,))
    if new_dbpath is None:
        new_dbpath = make_new_dbpath(ibs, 'aid', aid_list)
    gid_list = ut.unique(ibs.get_annot_gids(aid_list))
    nid_list = ut.unique(ibs.get_annot_nids(aid_list))
    return export_data(ibs, gid_list, aid_list, nid_list, new_dbpath=new_dbpath)


def export_data(ibs, gid_list, aid_list, nid_list, new_dbpath=None):
    """
    exports a subset of data and other required info

    Args:
        ibs (IBEISController):  wbia controller object
        gid_list (list):  list of image rowids
        aid_list (list):  list of annotation rowids
        nid_list (list):  list of name rowids
        imgsetid_list (list):  list of imageset rowids
        gsgrid_list (list):  list of imageset-image pairs rowids
        new_dbpath (None): (default = None)

    Returns:
        str: new_dbpath
    """
    import wbia

    imgsetid_list = ut.unique_unordered(ut.flatten(ibs.get_image_imgsetids(gid_list)))
    gsgrid_list = ut.unique_unordered(ut.flatten(ibs.get_image_gsgrids(gid_list)))

    # TODO: write SQL query to do this
    am_rowids = ibs._get_all_annotmatch_rowids()
    flags1_list = [aid in set(aid_list) for aid in ibs.get_annotmatch_aid1(am_rowids)]
    flags2_list = [aid in set(aid_list) for aid in ibs.get_annotmatch_aid2(am_rowids)]
    flag_list = ut.and_lists(flags1_list, flags2_list)
    am_rowids = ut.compress(am_rowids, flag_list)
    # am_rowids = ibs.get_valid_aids(ibs.get_valid_aids())

    rowid_subsets = {
        const.ANNOTATION_TABLE: aid_list,
        const.NAME_TABLE: nid_list,
        const.IMAGE_TABLE: gid_list,
        const.ANNOTMATCH_TABLE: am_rowids,
        const.GSG_RELATION_TABLE: gsgrid_list,
        const.IMAGESET_TABLE: imgsetid_list,
    }
    ibs_dst = wbia.opendb(dbdir=new_dbpath, allow_newdir=True)
    # Main merge driver
    merge_databases(ibs, ibs_dst, rowid_subsets=rowid_subsets)
    logger.info('Exported to %r' % (new_dbpath,))
    return new_dbpath


def slow_merge_test():
    r"""
    CommandLine:
        python -m wbia.dbio.export_subset --test-slow_merge_test

    Example:
        >>> # SLOW_DOCTEST
        >>> from wbia.dbio.export_subset import *  # NOQA
        >>> result = slow_merge_test()
        >>> print(result)
    """
    from wbia.dbio import export_subset
    import wbia

    ibs1 = wbia.opendb('testdb2')
    ibs1.fix_invalid_annotmatches()
    ibs_dst = wbia.opendb(db='testdb_dst2', allow_newdir=True, delete_ibsdir=True)
    export_subset.merge_databases(ibs1, ibs_dst)
    # ibs_src = ibs1
    check_merge(ibs1, ibs_dst)

    ibs2 = wbia.opendb('testdb1')
    ibs1.print_dbinfo()
    ibs2.print_dbinfo()
    ibs_dst.print_dbinfo()

    ibs_dst.print_dbinfo()

    export_subset.merge_databases(ibs2, ibs_dst)
    # ibs_src = ibs2
    check_merge(ibs2, ibs_dst)

    ibs3 = wbia.opendb('PZ_MTEST')
    export_subset.merge_databases(ibs3, ibs_dst)
    # ibs_src = ibs2
    check_merge(ibs3, ibs_dst)

    ibs_dst.print_dbinfo()
    return ibs_dst

    # ibs_src.print_annotation_table(exclude_columns=['annot_verts',
    # 'annot_semantic_uuid', 'annot_note', 'annot_parent_rowid',
    # 'annot_exemplar_flag,'])
    # ibs_dst.print_annotation_table()


def fix_bidirectional_annotmatch(ibs):
    import wbia

    infr = wbia.AnnotInference(ibs=ibs, aids='all', verbose=5)
    infr.initialize_graph()
    annots = ibs.annots()
    aid_to_nid = ut.dzip(annots.aids, annots.nids)

    # Delete bidirectional annotmatches
    annotmatch = ibs.db.get_table_as_pandas('annotmatch')
    df = annotmatch.set_index(['annot_rowid1', 'annot_rowid2'])

    # Find entires that have both directions
    pairs1 = annotmatch[['annot_rowid1', 'annot_rowid2']].values
    f_edges = {tuple(p) for p in pairs1}
    b_edges = {tuple(p[::-1]) for p in pairs1}
    isect_edges = {tuple(sorted(p)) for p in b_edges.intersection(f_edges)}
    logger.info('Found %d bidirectional edges' % len(isect_edges))
    isect_edges1 = list(isect_edges)
    isect_edges2 = [p[::-1] for p in isect_edges]

    import pandas as pd

    extra_ = {}
    fixme_edges = []
    d1 = df.loc[isect_edges1].reset_index(drop=False)
    d2 = df.loc[isect_edges2].reset_index(drop=False)
    flags = d1['annotmatch_evidence_decision'] != d2['annotmatch_evidence_decision']
    from wbia.tag_funcs import _parse_tags

    for f, r1, r2 in zip(flags, d1.iterrows(), d2.iterrows()):
        v1, v2 = r1[1], r2[1]
        aid1 = v1['annot_rowid1']
        aid2 = v1['annot_rowid2']
        truth_real = (
            ibs.const.EVIDENCE_DECISION.POSITIVE
            if aid_to_nid[aid1] == aid_to_nid[aid2]
            else ibs.const.EVIDENCE_DECISION.NEGATIVE
        )
        truth1 = v1['annotmatch_evidence_decision']
        truth2 = v2['annotmatch_evidence_decision']
        t1 = _parse_tags(v1['annotmatch_tag_text'])
        t2 = _parse_tags(v2['annotmatch_tag_text'])
        newtag = ut.union_ordered(t1, t2)
        fixme_flag = False
        if not pd.isnull(truth1):
            if truth_real != truth1:
                fixme_flag = True
        if not pd.isnull(truth2):
            if truth_real != truth2:
                fixme_flag = True
        if fixme_flag:
            logger.info('--')
            logger.info('t1, t2 = %r, %r' % (t1, t2))
            logger.info('newtag = %r' % (newtag,))
            logger.info(
                'truth_real, truth1, truth2 = %r, %r, %r' % (truth_real, truth1, truth2)
            )
            logger.info('aid1, aid2 = %r, %r' % (aid1, aid2))
            fixme_edges.append(tuple(sorted((aid1, aid2))))
        else:
            extra_[(aid1, aid2)] = (truth_real, newtag)

    if len(fixme_edges) > 0:
        # need to manually fix these edges
        fix_infr = wbia.AnnotInference.from_pairs(fixme_edges, ibs=ibs, verbose=5)
        feedback = fix_infr.read_wbia_annotmatch_feedback(only_existing_edges=True)
        infr = fix_infr

        fix_infr.external_feedback = feedback
        fix_infr.apply_feedback_edges()
        fix_infr.start_qt_interface(loop=False)
        # DELETE OLD EDGES TWICE
        ams = ibs.get_annotmatch_rowid_from_edges(fixme_edges)
        ibs.delete_annotmatch(ams)
        ams = ibs.get_annotmatch_rowid_from_edges(fixme_edges)
        ibs.delete_annotmatch(ams)

        # MANUALLY CALL THIS ONCE FINISHED
        # TO ONLY CHANGE ANNOTMATCH EDGES
        infr.write_wbia_staging_feedback()
        infr.write_wbia_annotmatch_feedback()

    # extra_.update(custom_)
    new_pairs = extra_.keys()
    new_truths = ut.take_column(ut.dict_take(extra_, new_pairs), 0)
    new_tags = ut.take_column(ut.dict_take(extra_, new_pairs), 1)
    new_tag_texts = [';'.join(t) for t in new_tags]
    aids1, aids2 = ut.listT(new_pairs)

    # Delete the old
    ibs.delete_annotmatch(
        (d1['annotmatch_rowid'].values.tolist() + d2['annotmatch_rowid'].values.tolist())
    )

    # Add the new
    ams = ibs.add_annotmatch_undirected(aids1, aids2)
    ibs.set_annotmatch_evidence_decision(ams, new_truths)
    ibs.set_annotmatch_tag_text(ams, new_tag_texts)

    if False:
        import wbia.guitool as gt

        gt.ensure_qapp()
        ut.qtensure()
        from wbia.gui import inspect_gui

        inspect_gui.show_vsone_tuner(ibs, aid1, aid2)


def fix_annotmatch_pzmaster1():
    """
    PZ_Master1 had annotmatch rowids that did not agree with the current name
    labeling. Looking at the inconsistencies in the graph interface was too
    cumbersome, because over 3000 annots were incorrectly grouped together.

    This function deletes any annotmatch rowid that is not consistent with the
    current labeling so we can go forward with using the new AnnotInference
    object
    """
    import wbia

    ibs = wbia.opendb('PZ_Master1')
    infr = wbia.AnnotInference(ibs=ibs, aids=ibs.get_valid_aids(), verbose=5)
    infr.initialize_graph()
    annots = ibs.annots()
    aid_to_nid = ut.dzip(annots.aids, annots.nids)

    if False:
        infr.reset_feedback()
        infr.ensure_mst()
        infr.apply_feedback_edges()
        infr.relabel_using_reviews()
        infr.start_qt_interface()

    # Get annotmatch rowids that agree with current labeling
    if False:
        annotmatch = ibs.db.get_table_as_pandas('annotmatch')
        import pandas as pd

        flags1 = pd.isnull(annotmatch['annotmatch_evidence_decision'])
        flags2 = annotmatch['annotmatch_tag_text'] == ''
        bad_part = annotmatch[flags1 & flags2]
        rowids = bad_part.index.tolist()
        ibs.delete_annotmatch(rowids)

    if False:
        # Delete bidirectional annotmatches
        annotmatch = ibs.db.get_table_as_pandas('annotmatch')
        df = annotmatch.set_index(['annot_rowid1', 'annot_rowid2'])

        # Find entires that have both directions
        pairs1 = annotmatch[['annot_rowid1', 'annot_rowid2']].values
        f_edges = {tuple(p) for p in pairs1}
        b_edges = {tuple(p[::-1]) for p in pairs1}
        isect_edges = {tuple(sorted(p)) for p in b_edges.intersection(f_edges)}
        isect_edges1 = list(isect_edges)
        isect_edges2 = [p[::-1] for p in isect_edges]

        # cols = ['annotmatch_evidence_decision', 'annotmatch_tag_text']
        import pandas as pd

        custom_ = {
            (559, 4909): (False, ['photobomb']),
            (7918, 8041): (False, ['photobomb']),
            (6634, 6754): (False, ['photobomb']),
            (3707, 3727): (False, ['photobomb']),
            (86, 103): (False, ['photobomb']),
        }
        extra_ = {}

        fixme_edges = []

        d1 = df.loc[isect_edges1].reset_index(drop=False)
        d2 = df.loc[isect_edges2].reset_index(drop=False)
        flags = d1['annotmatch_evidence_decision'] != d2['annotmatch_evidence_decision']
        from wbia.tag_funcs import _parse_tags

        for f, r1, r2 in zip(flags, d1.iterrows(), d2.iterrows()):
            v1, v2 = r1[1], r2[1]
            aid1 = v1['annot_rowid1']
            aid2 = v1['annot_rowid2']
            truth_real = (
                ibs.const.EVIDENCE_DECISION.POSITIVE
                if aid_to_nid[aid1] == aid_to_nid[aid2]
                else ibs.const.EVIDENCE_DECISION.NEGATIVE
            )
            truth1 = v1['annotmatch_evidence_decision']
            truth2 = v2['annotmatch_evidence_decision']
            t1 = _parse_tags(v1['annotmatch_tag_text'])
            t2 = _parse_tags(v2['annotmatch_tag_text'])
            newtag = ut.union_ordered(t1, t2)
            if (aid1, aid2) in custom_:
                continue
            fixme_flag = False
            if not pd.isnull(truth1):
                if truth_real != truth1:
                    fixme_flag = True
            if not pd.isnull(truth2):
                if truth_real != truth2:
                    fixme_flag = True
            if fixme_flag:
                logger.info('newtag = %r' % (newtag,))
                logger.info('truth_real = %r' % (truth_real,))
                logger.info('truth1 = %r' % (truth1,))
                logger.info('truth2 = %r' % (truth2,))
                logger.info('aid1 = %r' % (aid1,))
                logger.info('aid2 = %r' % (aid2,))
                fixme_edges.append((aid1, aid2))
            else:
                extra_[(aid1, aid2)] = (truth_real, newtag)

        extra_.update(custom_)
        new_pairs = extra_.keys()
        new_truths = ut.take_column(ut.dict_take(extra_, new_pairs), 0)
        new_tags = ut.take_column(ut.dict_take(extra_, new_pairs), 1)
        new_tag_texts = [';'.join(t) for t in new_tags]
        aids1, aids2 = ut.listT(new_pairs)

        # Delete the old
        ibs.delete_annotmatch(
            (
                d1['annotmatch_rowid'].values.tolist()
                + d2['annotmatch_rowid'].values.tolist()
            )
        )

        # Add the new
        ams = ibs.add_annotmatch_undirected(aids1, aids2)
        ibs.set_annotmatch_evidence_decision(ams, new_truths)
        ibs.set_annotmatch_tag_text(ams, new_tag_texts)

        if False:
            import wbia.guitool as gt

            gt.ensure_qapp()
            ut.qtensure()
            from wbia.gui import inspect_gui

            inspect_gui.show_vsone_tuner(ibs, aid1, aid2)

        # pairs2 = pairs1.T[::-1].T
        # idx1, idx2 = ut.isect_indices(list(map(tuple, pairs1)),
        #                               list(map(tuple, pairs2)))
        # r_edges = list(set(map(tuple, map(sorted, pairs1[idx1]))))
        # unique_pairs = list(set(map(tuple, map(sorted, pairs1[idx1]))))
        # df = annotmatch.set_index(['annot_rowid1', 'annot_rowid2'])

    x = ut.ddict(list)
    annotmatch = ibs.db.get_table_as_pandas('annotmatch')
    import ubelt as ub

    _iter = annotmatch.iterrows()
    prog = ub.ProgIter(_iter, length=len(annotmatch))
    for k, m in prog:
        aid1 = m['annot_rowid1']
        aid2 = m['annot_rowid2']
        if m['annotmatch_evidence_decision'] == ibs.const.EVIDENCE_DECISION.POSITIVE:
            if aid_to_nid[aid1] == aid_to_nid[aid2]:
                x['agree1'].append(k)
            else:
                x['disagree1'].append(k)
        elif m['annotmatch_evidence_decision'] == ibs.const.EVIDENCE_DECISION.NEGATIVE:
            if aid_to_nid[aid1] == aid_to_nid[aid2]:
                x['disagree2'].append(k)
            else:
                x['agree2'].append(k)

    ub.map_vals(len, x)
    ut.dict_hist(annotmatch.loc[x['disagree1']]['annotmatch_tag_text'])

    disagree1 = annotmatch.loc[x['disagree1']]
    pb_disagree1 = disagree1[disagree1['annotmatch_tag_text'] == 'photobomb']
    aids1 = pb_disagree1['annot_rowid1'].values.tolist()
    aids2 = pb_disagree1['annot_rowid2'].values.tolist()
    aid_pairs = list(zip(aids1, aids2))
    infr = wbia.AnnotInference.from_pairs(aid_pairs, ibs=ibs, verbose=5)
    if False:
        feedback = infr.read_wbia_annotmatch_feedback(edges=infr.edges())
        infr.external_feedback = feedback
        infr.apply_feedback_edges()
        infr.start_qt_interface(loop=False)

    # Delete these values
    if False:
        nonpb_disagree1 = disagree1[disagree1['annotmatch_tag_text'] != 'photobomb']
        disagree2 = annotmatch.loc[x['disagree2']]
        ibs.delete_annotmatch(nonpb_disagree1['annotmatch_rowid'])
        ibs.delete_annotmatch(disagree2['annotmatch_rowid'])

    # ut.dict_hist(disagree1['annotmatch_tag_text'])
    import networkx as nx

    graph = nx.Graph()
    graph.add_edges_from(zip(pb_disagree1['annot_rowid1'], pb_disagree1['annot_rowid2']))
    list(nx.connected_components(graph))

    set(annotmatch.loc[x['disagree2']]['annotmatch_tag_text'])

    # aid1, aid2 = 2585, 1875
    # # pd.unique(annotmatch['annotmatch_evidence_decision'])
    # from wbia.gui import inspect_gui
    # inspect_gui.show_vsone_tuner(ibs, aid1, aid2)
    # from vtool import inspect_matches

    # aid1, aid2 = 2108, 2040

    # pd.unique(annotmatch['annotmatch_tag_text'])

    # infr.reset_feedback()
    # infr.relabel_using_reviews()


def remerge_subset():
    """
    Assumes ibs1 is an updated subset of ibs2.
    Re-merges ibs1 back into ibs2.

    TODO: annotmatch table must have non-directional edges for this to work.
    I.e. u < v

    Ignore:

        # Ensure annotmatch and names are up to date with staging

        # Load graph
        import ibei
        ibs = wbia.opendb('PZ_PB_RF_TRAIN')
        infr = wbia.AnnotInference(aids='all', ibs=ibs, verbose=3)
        infr.reset_feedback('staging', apply=True)
        infr.relabel_using_reviews()

        # Check deltas
        infr.wbia_name_group_delta_info()
        infr.wbia_delta_info()

        # Write if it looks good
        infr.write_wbia_annotmatch_feedback()
        infr.write_wbia_name_assignment()

    Ignore:
        import wbia
        ibs = wbia.opendb('PZ_Master1')
        infr = wbia.AnnotInference(ibs, 'all')
        infr.reset_feedback('annotmatch', apply=True)

    CommandLine:
        python -m wbia.dbio.export_subset remerge_subset
    """
    import wbia

    ibs1 = wbia.opendb('PZ_PB_RF_TRAIN')
    ibs2 = wbia.opendb('PZ_Master1')

    gids1, gids2 = ibs1.images(), ibs2.images()
    idxs1, idxs2 = ut.isect_indices(gids1.uuids, gids2.uuids)
    isect_gids1, isect_gids2 = gids1.take(idxs1), gids2.take(idxs2)

    assert all(
        set.issubset(set(a1), set(a2))
        for a1, a2 in zip(isect_gids1.annot_uuids, isect_gids2.annot_uuids)
    )

    annot_uuids = ut.flatten(isect_gids1.annot_uuids)
    # aids1 = ibs1.annots(ibs1.get_annot_aids_from_uuid(annot_uuids), asarray=True)
    # aids2 = ibs2.annots(ibs2.get_annot_aids_from_uuid(annot_uuids), asarray=True)
    aids1 = ibs1.annots(uuids=annot_uuids, asarray=True)
    aids2 = ibs2.annots(uuids=annot_uuids, asarray=True)
    import numpy as np

    to_aids2 = dict(zip(aids1, aids2))
    # to_aids1 = dict(zip(aids2, aids1))

    # Step 1) Update individual annot properties
    # These annots need updates
    # np.where(aids1.visual_uuids != aids2.visual_uuids)
    # np.where(aids1.semantic_uuids != aids2.semantic_uuids)

    annot_unary_props = [
        # 'yaws', 'bboxes', 'thetas', 'qual', 'species', 'unary_tags']
        'yaws',
        'bboxes',
        'thetas',
        'qual',
        'species',
        'case_tags',
        'multiple',
        'age_months_est_max',
        'age_months_est_min',  # 'sex_texts'
    ]
    to_change = {}
    for key in annot_unary_props:
        prop1 = getattr(aids1, key)
        prop2 = getattr(aids2, key)
        diff_idxs = set(np.where(prop1 != prop2)[0])
        if diff_idxs:
            diff_prop1 = ut.take(prop1, diff_idxs)
            diff_prop2 = ut.take(prop2, diff_idxs)
            logger.info('key = %r' % (key,))
            logger.info('diff_prop1 = %r' % (diff_prop1,))
            logger.info('diff_prop2 = %r' % (diff_prop2,))
            to_change[key] = diff_idxs
    if to_change:
        changed_idxs = ut.unique(ut.flatten(to_change.values()))
        logger.info('Found %d annots that need updated properties' % len(changed_idxs))
        logger.info('changing unary attributes: %r' % (to_change,))
        if False and ut.are_you_sure('apply change'):
            for key, idxs in to_change.items():
                subaids1 = aids1.take(idxs)
                subaids2 = aids2.take(idxs)
                prop1 = getattr(subaids1, key)
                # prop2 = getattr(subaids2, key)
                setattr(subaids2, key, prop1)
    else:
        logger.info('Annot properties are in sync. Nothing to change')

    # Step 2) Update annotmatch - pairwise relationships
    infr1 = wbia.AnnotInference(aids=aids1.aids, ibs=ibs1, verbose=3, autoinit=False)

    # infr2 = wbia.AnnotInference(aids=ibs2.annots().aids, ibs=ibs2, verbose=3)
    aids2 = ibs2.get_valid_aids(is_known=True)
    infr2 = wbia.AnnotInference(aids=aids2, ibs=ibs2, verbose=3)
    infr2.reset_feedback('annotmatch', apply=True)

    # map feedback from ibs1 onto ibs2 using ibs2 aids.
    fb1 = infr1.read_wbia_annotmatch_feedback()
    fb1_t = {(to_aids2[u], to_aids2[v]): val for (u, v), val in fb1.items()}
    fb1_df_t = infr2._pandas_feedback_format(fb1_t).drop('am_rowid', axis=1)

    # Add transformed feedback into ibs2
    infr2.add_feedback_from(fb1_df_t)

    # Now ensure that dummy connectivity exists to preserve origninal names
    # from wbia.algo.graph import nx_utils
    # for (u, v) in infr2.find_mst_edges('name_label'):
    #     infr2.draw_aids((u, v))
    #     cc1 = infr2.pos_graph.connected_to(u)
    #     cc2 = infr2.pos_graph.connected_to(v)
    #     logger.info(nx_utils.edges_cross(infr2.graph, cc1, cc2))
    #     infr2.neg_redundancy(cc1, cc2)
    #     infr2.pos_redundancy(cc2)

    infr2.relabel_using_reviews(rectify=True)
    infr2.apply_nondynamic_update()

    if False:
        infr2.wbia_delta_info()
        infr2.wbia_name_group_delta_info()

    if len(list(infr2.inconsistent_components())) > 0:
        raise NotImplementedError('need to fix inconsistencies first')
        # Make it so it just loops until inconsistencies are resolved
        infr2.prioritize()
        infr2.qt_review_loop()
    else:
        infr2.write_wbia_staging_feedback()
        infr2.write_wbia_annotmatch_feedback()
        infr2.write_wbia_name_assignment()

    # if False:
    #     # Fix any inconsistency
    #     infr2.start_qt_interface(loop=False)
    #     test_nodes = [5344, 5430, 5349, 5334, 5383, 2280, 2265, 2234, 5399,
    #                   5338, 2654]
    #     import networkx as nx
    #     nx.is_connected(infr2.graph.subgraph(test_nodes))
    #     # infr = wbia.AnnotInference(aids=test_nodes, ibs=ibs2, verbose=5)

    #     # randomly sample some new labels to verify
    #     import wbia.guitool as gt
    #     from wbia.gui import inspect_gui
    #     gt.ensure_qapp()
    #     ut.qtensure()
    #     old_groups = ut.group_items(name_delta.index.tolist(), name_delta['old_name'])
    #     del old_groups['____']

    #     new_groups = ut.group_items(name_delta.index.tolist(), name_delta['new_name'])

    #     from wbia.algo.hots import simulate
    #     c = simulate.compare_groups(
    #         list(new_groups.values()),
    #         list(old_groups.values()),
    #     )
    #     ut.map_vals(len, c)
    #     for aids in c['pred_splits']:
    #         old_nids = ibs2.get_annot_nids(aids)
    #         new_nids = ut.take_column(infr2.gen_node_attrs('name_label', aids), 1)
    #         split_aids = ut.take_column(ut.group_items(aids, new_nids).values(), 0)
    #         aid1, aid2 = split_aids[0:2]

    #         if False:
    #             inspect_gui.show_vsone_tuner(ibs2, aid1, aid2)
    #     infr2.start_qt_interface(loop=False)

    # if False:
    #     # import wbia
    #     ibs1 = wbia.opendb('PZ_PB_RF_TRAIN')
    #     infr1 = wbia.AnnotInference(aids='all', ibs=ibs1, verbose=3)
    #     infr1.initialize_graph()
    #     # infr1.reset_feedback('staging')
    #     infr1.reset_feedback('annotmatch')
    #     infr1.apply_feedback_edges()
    #     infr1.relabel_using_reviews()
    #     infr1.apply_review_inference()
    #     infr1.start_qt_interface(loop=False)
    # delta = infr2.match_state_delta()
    # logger.info('delta = %r' % (delta,))

    # infr2.ensure_mst()
    # infr2.relabel_using_reviews()
    # infr2.apply_review_inference()

    # mst_edges = infr2.find_mst_edges()
    # set(infr2.graph.edges()).intersection(mst_edges)

    return
    """
    TODO:
        Task 2:
            Build AnnotInfr for ibs2 then add all decision from
            ibs1 to the internal feedback dict.

            Ensure that all other (esp old name-id related) edges are correctly
            placed, then overrite with new vals (
                make sure implicit vals do not cuase conflicts with new
                explicit vals, but old explicit vals should cause a conflict).
            Then just commit to staging and then commit to annotmatch and
            re-infer the names.
    """

    # Print some info about the delta
    # def _to_tup(x):
    #     return tuple(x) if isinstance(x, list) else x
    # changetype_list = list(zip(
    #     delta['old_decision'], delta['new_decision'],
    #     map(_to_tup, delta['old_tags']),
    #     map(_to_tup, delta['new_tags'])))
    # changetype_hist = ut.dict_hist(changetype_list, ordered=True)
    # logger.info(ut.align(ut.repr4(changetype_hist), ':'))

    # import pandas as pd
    # pd.options.display.max_rows = 20
    # pd.options.display.max_columns = 40
    # pd.options.display.width = 160
    # pd.options.display.float_format = lambda x: '%.4f' % (x,)

    # a, b = 86,    6265
    # c, d = to_aids1[a], to_aids1[b]
    # inspect_gui.show_vsone_tuner(ibs2, a, b)
    # inspect_gui.show_vsone_tuner(ibs1, to_aids1[a], to_aids1[b])
    # am1 = ibs1.get_annotmatch_rowids_between([to_aids1[a]],
    #                                          [to_aids1[b]])
    # am2 = ibs2.get_annotmatch_rowids_between([a], [b])
    # logger.info(ibs1.db.get_table_csv('annotmatch', rowids=am1))
    # logger.info(ibs2.db.get_table_csv('annotmatch', rowids=am2))

    # inspect_gui.show_vsone_tuner(ibs2, 8, 242)
    # inspect_gui.show_vsone_tuner(ibs2, 86, 103)
    # inspect_gui.show_vsone_tuner(ibs2, 86, 6265)


def find_overlap_annots(ibs1, ibs2, method='annots'):
    """
    Finds the aids of annotations in ibs1 that are also in ibs2

    ibs1 = wbia.opendb('PZ_Master1')
    ibs2 = wbia.opendb('PZ_MTEST')
    """
    if method == 'images':
        images1, images2 = ibs1.images(), ibs2.images()
        idxs1, idxs2 = ut.isect_indices(images1.uuids, images2.uuids)
        isect_images1 = images1.take(idxs1)
        annot_uuids = ut.flatten(isect_images1.annot_uuids)
        isect_annots1 = ibs1.annots(uuids=annot_uuids)
    elif method == 'annots':
        annots1, annots2 = ibs1.annots(), ibs2.annots()
        idxs1, idxs2 = ut.isect_indices(annots1.uuids, annots2.uuids)
        isect_annots1 = annots1.take(idxs1)
    return isect_annots1.aids


def check_database_overlap(ibs1, ibs2):
    """
    CommandLine:
        python -m wbia.other.dbinfo --test-get_dbinfo:1 --db PZ_MTEST
        dev.py -t listdbs
        python -m wbia.dbio.export_subset check_database_overlap
        --db PZ_MTEST --db2 PZ_MOTHERS

    CommandLine:
        python -m wbia.dbio.export_subset check_database_overlap

        python -m wbia.dbio.export_subset check_database_overlap --db1=PZ_MTEST --db2=PZ_Master0  # NOQA
        python -m wbia.dbio.export_subset check_database_overlap --db1=NNP_Master3 --db2=PZ_Master0  # NOQA

        python -m wbia.dbio.export_subset check_database_overlap --db1=GZ_Master0 --db2=GZ_ALL
        python -m wbia.dbio.export_subset check_database_overlap --db1=GZ_ALL --db2=lewa_grevys

        python -m wbia.dbio.export_subset check_database_overlap --db1=PZ_FlankHack --db2=PZ_Master1
        python -m wbia.dbio.export_subset check_database_overlap --db1=PZ_PB_RF_TRAIN --db2=PZ_Master1


    Example:
        >>> # SCRIPT
        >>> from wbia.dbio.export_subset import *  # NOQA
        >>> import wbia
        >>> import utool as ut
        >>> #ibs1 = wbia.opendb(db='PZ_Master0')
        >>> #ibs2 = wbia.opendb(dbdir='/raid/work2/Turk/PZ_Master')
        >>> db1 = ut.get_argval('--db1', str, default='PZ_MTEST')
        >>> db2 = ut.get_argval('--db2', str, default='testdb1')
        >>> dbdir1 = ut.get_argval('--dbdir1', str, default=None)
        >>> dbdir2 = ut.get_argval('--dbdir2', str, default=None)
        >>> ibs1 = wbia.opendb(db=db1, dbdir=dbdir1)
        >>> ibs2 = wbia.opendb(db=db2, dbdir=dbdir2)
        >>> check_database_overlap(ibs1, ibs2)
    """
    import numpy as np

    def print_isect(items1, items2, lbl=''):
        set1_ = set(items1)
        set2_ = set(items2)
        items_isect = set1_.intersection(set2_)
        fmtkw1 = dict(
            part=1,
            lbl=lbl,
            num=len(set1_),
            num_isect=len(items_isect),
            percent=100 * len(items_isect) / len(set1_),
        )
        fmtkw2 = dict(
            part=2,
            lbl=lbl,
            num=len(set2_),
            num_isect=len(items_isect),
            percent=100 * len(items_isect) / len(set2_),
        )
        fmt_a = '  * Num {lbl} {part}: {num_isect} / {num} = {percent:.2f}%'
        # fmt_b = '  * Num {lbl} isect: {num}'
        logger.info('Checking {lbl} intersection'.format(lbl=lbl))
        logger.info(fmt_a.format(**fmtkw1))
        logger.info(fmt_a.format(**fmtkw2))
        # logger.info(fmt_b.format(lbl=lbl, num=len(items_isect)))
        # items = items_isect
        # list_ = items1
        x_list1 = ut.find_list_indexes(items1, items_isect)
        x_list2 = ut.find_list_indexes(items2, items_isect)
        return x_list1, x_list2

    gids1 = ibs1.images()
    gids2 = ibs2.images()

    # Find common images
    # items1, items2, lbl, = gids1.uuids, gids2.uuids, 'images'
    gx_list1, gx_list2 = print_isect(gids1.uuids, gids2.uuids, 'images')
    gids_isect1 = gids1.take(gx_list1)
    gids_isect2 = gids2.take(gx_list2)
    assert gids_isect2.uuids == gids_isect1.uuids, 'sequence must be aligned'

    SHOW_ISECT_GIDS = False
    if SHOW_ISECT_GIDS:
        if len(gx_list1) > 0:
            logger.info('gids_isect1 = %r' % (gids_isect1,))
            logger.info('gids_isect2 = %r' % (gids_isect2,))
            if False:
                # Debug code
                import wbia.viz
                import wbia.plottool as pt

                gid_pairs = list(zip(gids_isect1, gids_isect2))
                pairs_iter = ut.ichunks(gid_pairs, chunksize=8)
                for fnum, pairs in enumerate(pairs_iter, start=1):
                    pnum_ = pt.make_pnum_nextgen(nRows=len(pairs), nCols=2)
                    for gid1, gid2 in pairs:
                        wbia.viz.show_image(ibs1, gid1, pnum=pnum_(), fnum=fnum)
                        wbia.viz.show_image(ibs2, gid2, pnum=pnum_(), fnum=fnum)

    # if False:
    #     aids1 = ibs1.get_valid_aids()
    #     aids2 = ibs2.get_valid_aids()
    #     ibs1.update_annot_visual_uuids(aids1)
    #     ibs2.update_annot_visual_uuids(aids2)
    #     ibs1.update_annot_semantic_uuids(aids1)
    #     ibs2.update_annot_semantic_uuids(aids2)

    # Check to see which intersecting images have different annotations
    image_aids_isect1 = gids_isect1.aids
    image_aids_isect2 = gids_isect2.aids
    image_avuuids_isect1 = np.array(
        ibs1.unflat_map(ibs1.get_annot_visual_uuids, image_aids_isect1)
    )
    image_avuuids_isect2 = np.array(
        ibs2.unflat_map(ibs2.get_annot_visual_uuids, image_aids_isect2)
    )
    changed_image_xs = np.nonzero(image_avuuids_isect1 != image_avuuids_isect2)[0]
    if len(changed_image_xs) > 0:
        logger.info(
            'There are %d images with changes in annotation visual information'
            % (len(changed_image_xs),)
        )
        changed_gids1 = ut.take(gids_isect1, changed_image_xs)
        changed_gids2 = ut.take(gids_isect2, changed_image_xs)

        SHOW_CHANGED_GIDS = False
        if SHOW_CHANGED_GIDS:
            logger.info('gids_isect1 = %r' % (changed_gids2,))
            logger.info('gids_isect2 = %r' % (changed_gids1,))
            # if False:
            #     # Debug code
            #     import wbia.viz
            #     import wbia.plottool as pt
            #     gid_pairs = list(zip(changed_gids1, changed_gids2))
            #     pairs_iter = ut.ichunks(gid_pairs, chunksize=8)
            #     for fnum, pairs in enumerate(pairs_iter, start=1):
            #         pnum_ = pt.make_pnum_nextgen(nRows=len(pairs), nCols=2)
            #         for gid1, gid2 in pairs:
            #             wbia.viz.show_image(
            #                 ibs1, gid1, pnum=pnum_(), fnum=fnum)
            #             wbia.viz.show_image(
            #                 ibs2, gid2, pnum=pnum_(), fnum=fnum)

    # Check for overlapping annotations (visual info only) in general
    aids1 = ibs1.annots()
    aids2 = ibs2.annots()

    # Check for overlapping annotations (visual + semantic info) in general
    aux_list1, aux_list2 = print_isect(aids1.uuids, aids2.uuids, 'uuids')
    avx_list1, avx_list2 = print_isect(aids1.visual_uuids, aids2.visual_uuids, 'vuuids')
    asx_list1, asx_list2 = print_isect(
        aids1.semantic_uuids, aids2.semantic_uuids, 'suuids'
    )

    # Check which images with the same visual uuids have different semantic
    # uuids
    changed_ax_list1 = ut.setdiff_ordered(avx_list1, asx_list1)
    changed_ax_list2 = ut.setdiff_ordered(avx_list2, asx_list2)
    assert len(changed_ax_list1) == len(changed_ax_list2)
    assert ut.take(aids1.visual_uuids, changed_ax_list1) == ut.take(
        aids2.visual_uuids, changed_ax_list2
    )

    changed_aids1 = np.array(ut.take(aids1, changed_ax_list1))
    changed_aids2 = np.array(ut.take(aids2, changed_ax_list2))

    changed_sinfo1 = ibs1.get_annot_semantic_uuid_info(changed_aids1)
    changed_sinfo2 = ibs2.get_annot_semantic_uuid_info(changed_aids2)
    sinfo1_arr = np.array(changed_sinfo1)
    sinfo2_arr = np.array(changed_sinfo2)
    is_semantic_diff = sinfo2_arr != sinfo1_arr
    # Inspect semantic differences
    if np.any(is_semantic_diff):
        colxs, rowxs = np.nonzero(is_semantic_diff)
        colx2_rowids = ut.group_items(rowxs, colxs)
        prop2_rowids = ut.map_dict_keys(changed_sinfo1._fields.__getitem__, colx2_rowids)
        logger.info(
            'changed_value_counts = ' + ut.repr2(ut.map_dict_vals(len, prop2_rowids))
        )
        yawx = changed_sinfo1._fields.index('yaw')

        # Show change in viewpoints
        if len(colx2_rowids[yawx]) > 0:
            vp_category_diff = ibsfuncs.viewpoint_diff(
                sinfo1_arr[yawx], sinfo2_arr[yawx]
            ).astype(np.float)
            # Look for category changes
            # any_diff = np.floor(vp_category_diff) > 0
            # _xs    = np.nonzero(any_diff)[0]
            # _aids1 = changed_aids1.take(_xs)
            # _aids2 = changed_aids2.take(_xs)
            # Look for significant changes
            is_significant_diff = np.floor(vp_category_diff) > 1
            significant_xs = np.nonzero(is_significant_diff)[0]
            significant_aids1 = changed_aids1.take(significant_xs)
            significant_aids2 = changed_aids2.take(significant_xs)
            logger.info(
                'There are %d significant viewpoint changes' % (len(significant_aids2),)
            )
            # vt.ori_distance(sinfo1_arr[yawx], sinfo2_arr[yawx])
            # zip(ibs1.get_annot_viewpoint_code(significant_aids1),
            # ibs2.get_annot_viewpoint_code(significant_aids2))
            # logger.info('yawdiff = %r' % )
            # if False:
            # Hack: Apply fixes
            # good_yaws = ibs2.get_annot_yaws(significant_aids2)
            # ibs1.set_annot_yaws(significant_aids1, good_yaws)
            #    pass
            if False:
                # Debug code
                import wbia.viz
                import wbia.plottool as pt

                # aid_pairs = list(zip(_aids1, _aids2))
                aid_pairs = list(zip(significant_aids1, significant_aids2))
                pairs_iter = ut.ichunks(aid_pairs, chunksize=8)
                for fnum, pairs in enumerate(pairs_iter, start=1):
                    pnum_ = pt.make_pnum_nextgen(nRows=len(pairs), nCols=2)
                    for aid1, aid2 in pairs:
                        wbia.viz.show_chip(
                            ibs1,
                            aid1,
                            pnum=pnum_(),
                            fnum=fnum,
                            show_viewcode=True,
                            nokpts=True,
                        )
                        wbia.viz.show_chip(
                            ibs2,
                            aid2,
                            pnum=pnum_(),
                            fnum=fnum,
                            show_viewcode=True,
                            nokpts=True,
                        )

    #
    nAnnots_per_image1 = np.array(ibs1.get_image_num_annotations(gids1))
    nAnnots_per_image2 = np.array(ibs2.get_image_num_annotations(gids2))
    #
    images_without_annots1 = sum(nAnnots_per_image1 == 0)
    images_without_annots2 = sum(nAnnots_per_image2 == 0)
    logger.info('images_without_annots1 = %r' % (images_without_annots1,))
    logger.info('images_without_annots2 = %r' % (images_without_annots2,))

    nAnnots_per_image1


"""
def MERGE_NNP_MASTER_SCRIPT():
    logger.info(ut.truncate_str(ibs_dst.db.get_table_csv(wbia.const.ANNOTATION_TABLE,
        exclude_columns=['annot_verts', 'annot_semantic_uuid', 'annot_note', 'annot_parent_rowid']), 10000))
    logger.info(ut.truncate_str(ibs_src1.db.get_table_csv(wbia.const.ANNOTATION_TABLE,
        exclude_columns=['annot_verts', 'annot_semantic_uuid', 'annot_note', 'annot_parent_rowid']), 10000))
    logger.info(ut.truncate_str(ibs_src1.db.get_table_csv(wbia.const.ANNOTATION_TABLE), 10000))

    from wbia.dbio.export_subset import *  # NOQA
    import wbia
    # Step 1
    ibs_src1 = wbia.opendb('GZC')
    ibs_dst = wbia.opendb('NNP_Master3', allow_newdir=True)
    merge_databases(ibs_src1, ibs_dst)

    ibs_src2 = wbia.opendb('NNP_initial')
    merge_databases(ibs_src2, ibs_dst)

    ## Step 2
    #ibs_src = wbia.opendb('GZC')

    ## Check
    ibs1 = wbia.opendb('NNP_initial')
    ibs2 = wbia.opendb('GZC')
    ibs3 = ibs_dst

    #logger.info(ibs1.get_image_time_statstr())
    #logger.info(ibs2.get_image_time_statstr())
    #logger.info(ibs3.get_image_time_statstr())
"""



# -*- coding: utf-8 -*-
import logging
import utool as ut
import six

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class AlignedListDictProxy(ut.DictLike_old):
    """
    simulates a dict when using parallel lists the point of this class is that
    when there are many instances of this class, then key2_idx can be shared between
    them. Ideally this class wont be used and will disappear when the parallel
    lists are being used properly.

    DEPCIRATE AlignedListDictProxy's defaultdict behavior is weird
    """

    def __init__(self, key2_idx, key_list, val_list):
        # if isinstance(key_list, np.ndarray):
        #    key_list = key_list.tolist()
        self.key_list = key_list
        self.val_list = val_list
        self.key2_idx = key2_idx
        self.default_function = None

    def __eq__(self, key):
        raise NotImplementedError()

    def pop(self, key):
        raise NotImplementedError()

    def __getitem__(self, key):
        try:
            idx = self.key2_idx[key]
        except (KeyError, IndexError):
            if self.default_function is not None:
                # behave like a default dict here
                self[key] = self.default_function()
                return self[key]
            else:
                raise
        return self.val_list[idx]
        # return ut.take(self.val_list, ut.dict_take(self.key2_idx, key))

    def __setitem__(self, key, val):
        try:
            idx = self.key2_idx[key]
        except KeyError:
            idx = len(self.key_list)
            self.key_list.append(key)
            self.key2_idx[key] = idx
        try:
            self.val_list[idx] = val
        except IndexError:
            if idx == len(self.val_list):
                self.val_list.append(val)
            else:
                raise
            # else:
            #    offset = idx - len(self.val_list)
            #    self.val_list.extend(([None] * offset) + [val])

    def iteritems(self):
        for key, val in zip(self.key_list, self.val_list):
            yield key, val

    def iterkeys(self):
        return iter(self.key_list)

    def itervalues(self):
        return iter(self.val_list)


class _OldStyleChipMatchSimulator(object):
    # SIMULATE OLD CHIPMATCHES UNTIL TRANSFER IS COMPLETE
    # TRY NOT TO USE THESE AS THEY WILL BE MUCH SLOWER THAN
    # NORMAL.
    _oldfields = ('aid2_fm', 'aid2_fsv', 'aid2_fk', 'aid2_score', 'aid2_H')

    def __iter__(cm):
        for field in cm._oldfields:
            yield getattr(cm, field)

    def __getitem__(cm, index):
        if isinstance(index, six.string_types):
            return cm.__dict__[index]
        else:
            return getattr(cm, cm._oldfields[index])

    @property
    def aid2_fm(cm):
        return AlignedListDictProxy(cm.daid2_idx, cm.daid_list, cm.fm_list)

    @property
    def aid2_fsv(cm):
        return AlignedListDictProxy(cm.daid2_idx, cm.daid_list, cm.fsv_list)

    @property
    def aid2_fk(cm):
        return AlignedListDictProxy(cm.daid2_idx, cm.daid_list, cm.fk_list)

    @property
    def aid2_H(cm):
        return (
            None
            if cm.H_list is None
            else AlignedListDictProxy(cm.daid2_idx, cm.daid_list, cm.H_list)
        )

    @property
    def aid2_score(cm):
        return (
            {}
            if cm.score_list is None
            else AlignedListDictProxy(cm.daid2_idx, cm.daid_list, cm.score_list)
        )

    @property
    def aid2_fs(cm):
        if cm.fs_list is None:
            fs_list = cm.get_fsv_prod_list()
        else:
            fs_list = cm.fs_list
        return AlignedListDictProxy(cm.daid2_idx, cm.daid_list, fs_list)

    # qres compatibility

    # @property
    # def filtkey_list(cm):
    #     """ for compatibility with qres """
    #     return cm.fsv_col_lbls

    # @property
    # def nid2_name_score(cm):
    #     """
    #     DEPCIRATE AlignedListDictProxy's defaultdict behavior is weird

    #     """
    #     return ({} if cm.score_list is None else
    #             AlignedListDictProxy(cm.nid2_nidx, cm.unique_nids, cm.name_score_list))

    # @property
    # def aid2_annot_score(cm):
    #     """ DEPCIRATE AlignedListDictProxy's defaultdict behavior is weird

    #     ute cm.get_annot_scores instead

    #     """
    #     return ({} if cm.annot_score_list is None else
    #             AlignedListDictProxy(cm.daid2_idx, cm.daid_list, cm.annot_score_list))

    # def get_nscoretup(cm):
    #     return cm.get_ranked_nids_and_aids()

    # def tokwargs(cm):
    #     """
    #     Can be unpacked and passed as kwargs
    #     **cm.tokwargs()
    #     """
    #     return ut.KwargsWrapper(cm)
